'''
This is a variable hierarchy that is for a single mission. Each mission
gets a copy of this hierarchy.
'''


# ---------------------------
# Aircraft data hierarchy
# ---------------------------
class Aircraft:

    class AirConditioning:
        MASS = 'aircraft:air_conditioning:mass'
        MASS_SCALER = 'aircraft:air_conditioning:mass_scaler'

    class AntiIcing:
        MASS = 'aircraft:anti_icing:mass'
        MASS_SCALER = 'aircraft:anti_icing:mass_scaler'

    class APU:
        MASS = 'aircraft:apu:mass'
        MASS_SCALER = 'aircraft:apu:mass_scaler'

    class Avionics:
        MASS = 'aircraft:avionics:mass'
        MASS_SCALER = 'aircraft:avionics:mass_scaler'

    class BWB:
        CABIN_AREA = 'aircraft:blended_wing_body_design:cabin_area'
        NUM_BAYS = 'aircraft:blended_wing_body_design:num_bays'

        PASSENGER_LEADING_EDGE_SWEEP = \
            'aircraft:blended_wing_body_design:passenger_leading_edge_sweep'

    class Canard:
        AREA = 'aircraft:canard:area'
        ASPECT_RATIO = 'aircraft:canard:aspect_ratio'
        CHARACTERISTIC_LENGTH = 'aircraft:canard:characteristic_length'
        FINENESS = 'aircraft:canard:fineness'
        LAMINAR_FLOW_LOWER = 'aircraft:canard:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:canard:laminar_flow_upper'
        MASS = 'aircraft:canard:mass'
        MASS_SCALER = 'aircraft:canard:mass_scaler'
        TAPER_RATIO = 'aircraft:canard:taper_ratio'
        THICKNESS_TO_CHORD = 'aircraft:canard:thickness_to_chord'
        WETTED_AREA = 'aircraft:canard:wetted_area'
        WETTED_AREA_SCALER = 'aircraft:canard:wetted_area_scaler'

    class Controls:
        COCKPIT_CONTROL_MASS_SCALER = 'aircraft:controls:cockpit_control_mass_scaler'
        CONTROL_MASS_INCREMENT = 'aircraft:controls:control_mass_increment'
        STABILITY_AUGMENTATION_SYSTEM_MASS = 'aircraft:controls:stability_augmentation_system_mass'
        STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER = 'aircraft:controls:stability_augmentation_system_mass_scaler'
        TOTAL_MASS = 'aircraft:controls:total_mass'

    class CrewPayload:
        BAGGAGE_MASS = 'aircraft:crew_and_payload:baggage_mass'

        BAGGAGE_MASS_PER_PASSENGER = \
            'aircraft:crew_and_payload:baggage_mass_per_passenger'

        CARGO_CONTAINER_MASS = \
            'aircraft:crew_and_payload:cargo_container_mass'

        CARGO_CONTAINER_MASS_SCALER = \
            'aircraft:crew_and_payload:cargo_container_mass_scaler'

        CARGO_MASS = 'aircraft:crew_and_payload:cargo_mass'
        FLIGHT_CREW_MASS = 'aircraft:crew_and_payload:flight_crew_mass'

        FLIGHT_CREW_MASS_SCALER = \
            'aircraft:crew_and_payload:flight_crew_mass_scaler'

        MASS_PER_PASSENGER = 'aircraft:crew_and_payload:mass_per_passenger'

        MISC_CARGO = 'aircraft:crew_and_payload:misc_cargo'

        NON_FLIGHT_CREW_MASS = \
            'aircraft:crew_and_payload:non_flight_crew_mass'

        NON_FLIGHT_CREW_MASS_SCALER = \
            'aircraft:crew_and_payload:non_flight_crew_mass_scaler'

        NUM_BUSINESS_CLASS = 'aircraft:crew_and_payload:num_business_class'
        NUM_FIRST_CLASS = 'aircraft:crew_and_payload:num_first_class'

        NUM_FLIGHT_ATTENDANTS = \
            'aircraft:crew_and_payload:num_flight_attendants'

        NUM_FLIGHT_CREW = 'aircraft:crew_and_payload:num_flight_crew'
        NUM_GALLEY_CREW = 'aircraft:crew_and_payload:num_galley_crew'

        NUM_PASSENGERS = 'aircraft:crew_and_payload:num_passengers'
        NUM_TOURIST_CLASS = 'aircraft:crew_and_payload:num_tourist_class'

        PASSENGER_MASS = \
            'aircraft:crew_and_payload:passenger_mass'
        PASSENGER_MASS_WITH_BAGS = \
            'aircraft:crew_and_payload:passenger_mass_with_bags'

        PASSENGER_PAYLOAD_MASS = 'aircraft:crew_and_payload:passenger_payload_mass'

        PASSENGER_SERVICE_MASS = \
            'aircraft:crew_and_payload:passenger_service_mass'

        PASSENGER_SERVICE_MASS_SCALER = \
            'aircraft:crew_and_payload:passenger_service_mass_scaler'

        TOTAL_PAYLOAD_MASS = 'aircraft:crew_and_payload:total_payload_mass'
        WING_CARGO = 'aircraft:crew_and_payload:wing_cargo'

    class Design:
        # These variables are values that do not fall into a particular aircraft
        # component.
        BASE_AREA = 'aircraft:design:base_area'
        CG_DELTA = 'aircraft:design:cg_delta'
        CHARACTERISTIC_LENGTHS = 'aircraft:design:characteristic_lengths'
        COCKPIT_CONTROL_MASS_COEFFICIENT = 'aircraft:design:cockpit_control_mass_coefficient'
        COMPUTE_HTAIL_VOLUME_COEFF = 'aircraft:design:compute_htail_volume_coeff'
        COMPUTE_VTAIL_VOLUME_COEFF = 'aircraft:design:compute_vtail_volume_coeff'
        DRAG_COEFFICIENT_INCREMENT = 'aircraft:design:drag_increment'
        DRAG_POLAR = 'aircraft:design:drag_polar'

        EMPTY_MASS = 'aircraft:design:empty_mass'
        EMPTY_MASS_MARGIN = 'aircraft:design:empty_mass_margin'

        EMPTY_MASS_MARGIN_SCALER = \
            'aircraft:design:empty_mass_margin_scaler'

        EQUIPMENT_MASS_COEFFICIENTS = 'aircraft:design:equipment_mass_coefficients'
        EXTERNAL_SUBSYSTEMS_MASS = 'aircraft:design:external_subsystems_mass'
        FINENESS = 'aircraft:design:fineness'
        FIXED_EQUIPMENT_MASS = 'aircraft:design:fixed_equipment_mass'
        FIXED_USEFUL_LOAD = 'aircraft:design:fixed_useful_load'
        IJEFF = 'ijeff'
        LAMINAR_FLOW_LOWER = 'aircraft:design:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:design:laminar_flow_upper'

        LANDING_TO_TAKEOFF_MASS_RATIO = \
            'aircraft:design:landing_to_takeoff_mass_ratio'

        LIFT_CURVE_SLOPE = 'aircraft:design:lift_curve_slope'
        LIFT_DEPENDENT_DRAG_COEFF_FACTOR = \
            'aircraft:design:lift_dependent_drag_coeff_factor'
        LIFT_POLAR = 'aircraft:design:lift_polar'

        MAX_FUSELAGE_PITCH_ANGLE = 'aircraft:design:max_fuselage_pitch_angle'
        MAX_STRUCTURAL_SPEED = 'aircraft:design:max_structural_speed'
        OPERATING_MASS = 'aircraft:design:operating_mass'
        PART25_STRUCTURAL_CATEGORY = 'aircraft:design:part25_structural_category'
        RESERVE_FUEL_ADDITIONAL = 'aircraft:design:reserve_fuel_additional'
        RESERVE_FUEL_FRACTION = 'aircraft:design:reserve_fuel_fraction'
        SMOOTH_MASS_DISCONTINUITIES = 'aircraft:design:smooth_mass_discontinuities'
        STATIC_MARGIN = 'aircraft:design:static_margin'
        STRUCTURAL_MASS_INCREMENT = 'aircraft:design:structural_mass_increment'
        STRUCTURE_MASS = 'aircraft:design:structure_mass'

        SUBSONIC_DRAG_COEFF_FACTOR = \
            'aircraft:design:subsonic_drag_coeff_factor'

        SUPERCRITICAL_DIVERGENCE_SHIFT = 'aircraft:design:supercritical_drag_shift'

        SUPERSONIC_DRAG_COEFF_FACTOR = \
            'aircraft:design:supersonic_drag_coeff_factor'

        SYSTEMS_EQUIP_MASS = 'aircraft:design:systems_equip_mass'
        SYSTEMS_EQUIP_MASS_BASE = 'aircraft:design:systems_equip_mass_base'
        THRUST_TO_WEIGHT_RATIO = 'aircraft:design:thrust_to_weight_ratio'
        TOTAL_WETTED_AREA = 'aircraft:design:total_wetted_area'
        TOUCHDOWN_MASS = 'aircraft:design:touchdown_mass'
        ULF_CALCULATED_FROM_MANEUVER = 'aircraft:design:ulf_calculated_from_maneuver'
        USE_ALT_MASS = 'aircraft:design:use_alt_mass'
        WETTED_AREAS = 'aircraft:design:wetted_areas'
        ZERO_FUEL_MASS = 'aircraft:design:zero_fuel_mass'
        ZERO_LIFT_DRAG_COEFF_FACTOR = \
            'aircraft:design:zero_lift_drag_coeff_factor'

    class Electrical:
        HAS_HYBRID_SYSTEM = 'aircraft:electrical:has_hybrid_system'
        HYBRID_CABLE_LENGTH = 'aircraft:electrical:hybrid_cable_length'
        MASS = 'aircraft:electrical:mass'
        MASS_SCALER = 'aircraft:electrical:mass_scaler'

    class Engine:
        ADDITIONAL_MASS = \
            'aircraft:engine:additional_mass'

        ADDITIONAL_MASS_FRACTION = \
            'aircraft:engine:additional_mass_fraction'

        CONSTANT_FUEL_CONSUMPTION = 'aircraft:engine:constant_fuel_consumption'
        CONTROLS_MASS = 'aircraft:engine:controls_mass'

        DATA_FILE = 'aircraft:engine:data_file'
        FLIGHT_IDLE_MAX_FRACTION = 'aircraft:engine:flight_idle_max_fraction'
        FLIGHT_IDLE_MIN_FRACTION = 'aircraft:engine:flight_idle_min_fraction'
        FLIGHT_IDLE_THRUST_FRACTION = 'aircraft:engine:flight_idle_thrust_fraction'
        FUEL_FLOW_SCALER_CONSTANT_TERM = 'aircraft:engine:fuel_flow_scaler_constant_term'
        FUEL_FLOW_SCALER_LINEAR_TERM = 'aircraft:engine:fuel_flow_scaler_linear_term'
        GENERATE_FLIGHT_IDLE = 'aircraft:engine:generate_flight_idle'
        # GENERATE_STRUCTURED_DATA = 'aircraft:engine:generate_structured_data'
        GEOPOTENTIAL_ALT = 'aircraft:engine:geopotential_alt'
        HAS_PROPELLERS = 'aircraft:engine:has_propellers'
        IGNORE_NEGATIVE_THRUST = 'aircraft:engine:ignore_negative_thrust'
        INTERPOLATION_METHOD = 'aircraft:engine:interpolation_method'
        MASS = 'aircraft:engine:mass'
        MASS_SCALER = 'aircraft:engine:mass_scaler'
        MASS_SPECIFIC = 'aircraft:engine:mass_specific'
        MODEL_SLS_THRUST = 'aircraft:engine:model_sls_thrust'
        NUM_ENGINES = 'aircraft:engine:num_engines'
        NUM_FUSELAGE_ENGINES = 'aircraft:engine:num_fuselage_engines'
        NUM_WING_ENGINES = 'aircraft:engine:num_wing_engines'
        POD_MASS = 'aircraft:engine:pod_mass'
        POD_MASS_SCALER = 'aircraft:engine:pod_mass_scaler'
        POSITION_FACTOR = 'aircraft:engine:position_factor'
        PYLON_FACTOR = 'aircraft:engine:pylon_factor'
        REFERENCE_DIAMETER = 'aircraft:engine:reference_diameter'
        REFERENCE_MASS = 'aircraft:engine:reference_mass'
        REFERENCE_SLS_THRUST = 'aircraft:engine:reference_sls_thrust'
        SCALE_FACTOR = 'aircraft:engine:scale_factor'
        SCALE_MASS = 'aircraft:engine:scale_mass'
        SCALE_PERFORMANCE = 'aircraft:engine:scale_performance'
        SCALED_SLS_THRUST = 'aircraft:engine:scaled_sls_thrust'
        STARTER_MASS = 'aircraft:engine:starter_mass'
        SUBSONIC_FUEL_FLOW_SCALER = 'aircraft:engine:subsonic_fuel_flow_scaler'
        SUPERSONIC_FUEL_FLOW_SCALER = 'aircraft:engine:supersonic_fuel_flow_scaler'

        THRUST_REVERSERS_MASS = \
            'aircraft:engine:thrust_reversers_mass'

        THRUST_REVERSERS_MASS_SCALER = \
            'aircraft:engine:thrust_reversers_mass_scaler'

        TYPE = 'aircraft:engine:type'

        WING_LOCATIONS = \
            'aircraft:engine:wing_locations'

    class Fins:
        AREA = 'aircraft:fins:area'
        MASS = 'aircraft:fins:mass'
        MASS_SCALER = 'aircraft:fins:mass_scaler'
        NUM_FINS = 'aircraft:fins:num_fins'
        TAPER_RATIO = 'aircraft:fins:taper_ratio'

    class Fuel:
        AUXILIARY_FUEL_CAPACITY = 'aircraft:fuel:auxiliary_fuel_capacity'
        BURN_PER_PASSENGER_MILE = 'aircraft:fuel:burn_per_passenger_mile'
        CAPACITY_FACTOR = 'aircraft:fuel:capacity_factor'
        DENSITY = 'aircraft:fuel:density'
        DENSITY_RATIO = 'aircraft:fuel:density_ratio'
        FUEL_MARGIN = 'aircraft:fuel:fuel_margin'
        FUEL_SYSTEM_MASS = 'aircraft:fuel:fuel_system_mass'
        FUEL_SYSTEM_MASS_COEFFICIENT = 'aircraft:fuel:fuel_system_mass_coefficient'
        FUEL_SYSTEM_MASS_SCALER = 'aircraft:fuel:fuel_system_mass_scaler'
        FUSELAGE_FUEL_CAPACITY = 'aircraft:fuel:fuselage_fuel_capacity'
        NUM_TANKS = 'aircraft:fuel:num_tanks'
        TOTAL_CAPACITY = 'aircraft:fuel:total_capacity'
        TOTAL_VOLUME = 'aircraft:fuel:total_volume'
        UNUSABLE_FUEL_MASS = 'aircraft:fuel:unusable_fuel_mass'

        UNUSABLE_FUEL_MASS_SCALER = \
            'aircraft:fuel:unusable_fuel_mass_scaler'

        WING_FUEL_CAPACITY = 'aircraft:fuel:wing_fuel_capacity'
        WING_FUEL_FRACTION = 'aircraft:fuel:wing_fuel_fraction'
        WING_REF_CAPACITY = 'aircraft:fuel:wing_ref_capacity'
        WING_REF_CAPACITY_AREA = 'aircraft:fuel:wing_ref_capacity_area'
        WING_REF_CAPACITY_TERM_A = 'aircraft:fuel:wing_ref_capacity_term_A'
        WING_REF_CAPACITY_TERM_B = 'aircraft:fuel:wing_ref_capacity_term_B'
        # WING_VOLUME = 'aircraft:fuel:wing_volume'
        WING_VOLUME_DESIGN = 'aircraft:fuel:wing_volume_design'
        WING_VOLUME_GEOMETRIC_MAX = 'aircraft:fuel:wing_volume_geometric_max'
        WING_VOLUME_STRUCTURAL_MAX = 'aircraft:fuel:wing_volume_structural_max'

    class Furnishings:
        MASS = 'aircraft:furnishings:mass'
        MASS_BASE = 'aircraft:furnishings:mass_base'
        MASS_SCALER = 'aircraft:furnishings:mass_scaler'

    class Fuselage:
        AISLE_WIDTH = 'aircraft:fuselage:aisle_width'
        AVG_DIAMETER = 'aircraft:fuselage:avg_diameter'
        CHARACTERISTIC_LENGTH = 'aircraft:fuselage:characteristic_length'
        CROSS_SECTION = 'aircraft:fuselage:cross_section'
        DELTA_DIAMETER = 'aircraft:fuselage:delta_diameter'
        DIAMETER_TO_WING_SPAN = 'aircraft:fuselage:diameter_to_wing_span'
        FINENESS = 'aircraft:fuselage:fineness'
        FLAT_PLATE_AREA_INCREMENT = 'aircraft:fuselage:flat_plate_area_increment'
        FORM_FACTOR = 'aircraft:fuselage:form_factor'
        LAMINAR_FLOW_LOWER = 'aircraft:fuselage:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:fuselage:laminar_flow_upper'
        LENGTH = 'aircraft:fuselage:length'
        LENGTH_TO_DIAMETER = 'aircraft:fuselage:length_to_diameter'
        MASS = 'aircraft:fuselage:mass'
        MASS_COEFFICIENT = 'aircraft:fuselage:mass_coefficient'
        MASS_SCALER = 'aircraft:fuselage:mass_scaler'
        MAX_HEIGHT = 'aircraft:fuselage:max_height'
        MAX_WIDTH = 'aircraft:fuselage:max_width'
        MILITARY_CARGO_FLOOR = 'aircraft:fuselage:military_cargo_floor'
        NOSE_FINENESS = 'aircraft:fuselage:nose_fineness'
        NUM_AISLES = 'aircraft:fuselage:num_aisles'
        NUM_FUSELAGES = 'aircraft:fuselage:num_fuselages'
        NUM_SEATS_ABREAST = 'aircraft:fuselage:num_seats_abreast'

        PASSENGER_COMPARTMENT_LENGTH = \
            'aircraft:fuselage:passenger_compartment_length'

        PILOT_COMPARTMENT_LENGTH = 'aircraft:fuselage:pilot_compartment_length'
        PLANFORM_AREA = 'aircraft:fuselage:planform_area'
        PRESSURE_DIFFERENTIAL = 'aircraft:fuselage:pressure_differential'
        PROVIDE_SURFACE_AREA = 'aircraft:fuselage:provide_surface_area'
        SEAT_PITCH = 'aircraft:fuselage:seat_pitch'
        SEAT_WIDTH = 'aircraft:fuselage:seat_width'
        TAIL_FINENESS = 'aircraft:fuselage:tail_fineness'
        WETTED_AREA = 'aircraft:fuselage:wetted_area'
        WETTED_AREA_FACTOR = 'aircraft:fuselage:wetted_area_factor'
        WETTED_AREA_SCALER = 'aircraft:fuselage:wetted_area_scaler'

    class HorizontalTail:
        AREA = 'aircraft:horizontal_tail:area'
        ASPECT_RATIO = 'aircraft:horizontal_tail:aspect_ratio'
        AVERAGE_CHORD = 'aircraft:horizontal_tail:average_chord'

        CHARACTERISTIC_LENGTH = \
            'aircraft:horizontal_tail:characteristic_length'

        FINENESS = 'aircraft:horizontal_tail:fineness'
        FORM_FACTOR = 'aircraft:horizontal_tail:form_factor'
        LAMINAR_FLOW_LOWER = 'aircraft:horizontal_tail:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:horizontal_tail:laminar_flow_upper'
        MASS = 'aircraft:horizontal_tail:mass'
        MASS_COEFFICIENT = 'aircraft:horizontal_tail:mass_coefficient'
        MASS_SCALER = 'aircraft:horizontal_tail:mass_scaler'
        MOMENT_ARM = 'aircraft:horizontal_tail:moment_arm'
        MOMENT_RATIO = 'aircraft:horizontal_tail:moment_ratio'
        ROOT_CHORD = 'aircraft:horizontal_tail:root_chord'
        SPAN = 'aircraft:horizontal_tail:span'
        SWEEP = 'aircraft:horizontal_tail:sweep'
        TAPER_RATIO = 'aircraft:horizontal_tail:taper_ratio'
        THICKNESS_TO_CHORD = 'aircraft:horizontal_tail:thickness_to_chord'

        VERTICAL_TAIL_FRACTION = \
            'aircraft:horizontal_tail:vertical_tail_fraction'

        VOLUME_COEFFICIENT = 'aircraft:horizontal_tail:volume_coefficient'
        WETTED_AREA = 'aircraft:horizontal_tail:wetted_area'
        WETTED_AREA_SCALER = 'aircraft:horizontal_tail:wetted_area_scaler'

    class Hydraulics:
        MASS = 'aircraft:hydraulics:mass'
        MASS_SCALER = 'aircraft:hydraulics:mass_scaler'
        SYSTEM_PRESSURE = 'aircraft:hydraulics:system_pressure'

    class Instruments:
        MASS = 'aircraft:instruments:mass'
        MASS_SCALER = 'aircraft:instruments:mass_scaler'

    class LandingGear:
        CARRIER_BASED = 'aircraft:landing_gear:carrier_based'
        DRAG_COEFFICIENT = 'aircraft:landing_gear:drag_coefficient'
        FIXED_GEAR = 'aircraft:landing_gear:fixed_gear'
        MAIN_GEAR_LOCATION = 'aircraft:landing_gear:main_gear_location'
        MAIN_GEAR_MASS = 'aircraft:landing_gear:main_gear_mass'
        MAIN_GEAR_MASS_COEFFICIENT = 'aircraft:landing_gear:main_gear_mass_coefficient'
        MAIN_GEAR_MASS_SCALER = \
            'aircraft:landing_gear:main_gear_mass_scaler'
        MAIN_GEAR_OLEO_LENGTH = 'aircraft:landing_gear:main_gear_oleo_length'

        MASS_COEFFICIENT = 'aircraft:landing_gear:mass_coefficient'

        NOSE_GEAR_MASS = 'aircraft:landing_gear:nose_gear_mass'
        NOSE_GEAR_MASS_SCALER = \
            'aircraft:landing_gear:nose_gear_mass_scaler'
        NOSE_GEAR_OLEO_LENGTH = 'aircraft:landing_gear:nose_gear_oleo_length'

        TAIL_HOOK_MASS_SCALER = 'aircraft:landing_gear:tail_hook_mass_scaler'
        TOTAL_MASS = 'aircraft:landing_gear:total_mass'
        TOTAL_MASS_SCALER = 'aircraft:landing_gear:total_mass_scaler'

    class Nacelle:
        AVG_DIAMETER = 'aircraft:nacelle:avg_diameter'
        AVG_LENGTH = 'aircraft:nacelle:avg_length'
        CHARACTERISTIC_LENGTH = 'aircraft:nacelle:characteristic_length'
        CLEARANCE_RATIO = 'aircraft:nacelle:clearance_ratio'
        CORE_DIAMETER_RATIO = 'aircraft:nacelle:core_diameter_ratio'
        FINENESS = 'aircraft:nacelle:fineness'
        FORM_FACTOR = 'aircraft:nacelle:form_factor'
        LAMINAR_FLOW_LOWER = 'aircraft:nacelle:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:nacelle:laminar_flow_upper'
        MASS = 'aircraft:nacelle:mass'
        MASS_SCALER = 'aircraft:nacelle:mass_scaler'
        MASS_SPECIFIC = 'aircraft:nacelle:mass_specific'
        SURFACE_AREA = 'aircraft:nacelle:surface_area'
        TOTAL_WETTED_AREA = 'aircraft:nacelle:total_wetted_area'
        WETTED_AREA = 'aircraft:nacelle:wetted_area'
        WETTED_AREA_SCALER = 'aircraft:nacelle:wetted_area_scaler'

    class Paint:
        MASS = 'aircraft:paint:mass'
        MASS_PER_UNIT_AREA = 'aircraft:paint:mass_per_unit_area'

    class Propulsion:
        ENGINE_OIL_MASS_SCALER = \
            'aircraft:propulsion:engine_oil_mass_scaler'

        MASS = 'aircraft:propulsion:mass'
        MISC_MASS_SCALER = 'aircraft:propulsion:misc_mass_scaler'
        TOTAL_ENGINE_CONTROLS_MASS = 'aircraft:propulsion:total_engine_controls_mass'
        TOTAL_ENGINE_MASS = 'aircraft:propulsion:total_engine_mass'
        TOTAL_ENGINE_OIL_MASS = 'aircraft:propulsion:total_engine_oil_mass'
        TOTAL_ENGINE_POD_MASS = 'aircraft:propulsion:total_engine_pod_mass'
        TOTAL_MISC_MASS = 'aircraft:propulsion:total_misc_mass'
        TOTAL_NUM_ENGINES = 'aircraft:propulsion:total_num_engines'
        TOTAL_NUM_FUSELAGE_ENGINES = 'aircraft:propulsion:total_num_fuselage_engines'
        TOTAL_NUM_WING_ENGINES = 'aircraft:propulsion:total_num_wing_engines'
        TOTAL_REFERENCE_SLS_THRUST = 'aircraft:propulsion:total_reference_sls_thrust'
        TOTAL_SCALED_SLS_THRUST = 'aircraft:propulsion:total_scaled_sls_thrust'
        TOTAL_STARTER_MASS = 'aircraft:propulsion:total_starter_mass'

        TOTAL_THRUST_REVERSERS_MASS = \
            'aircraft:propulsion:total_thrust_reversers_mass'

    class Strut:
        AREA = 'aircraft:strut:area'
        AREA_RATIO = 'aircraft:strut:area_ratio'
        ATTACHMENT_LOCATION = 'aircraft:strut:attachment_location'
        ATTACHMENT_LOCATION_DIMENSIONLESS = 'aircraft:strut:attachment_location_dimensionless'
        CHORD = 'aircraft:strut:chord'
        DIMENSIONAL_LOCATION_SPECIFIED = 'aircraft:strut:dimensional_location_specified'
        FUSELAGE_INTERFERENCE_FACTOR = 'aircraft:strut:fuselage_interference_factor'
        LENGTH = 'aircraft:strut:length'
        MASS = 'aircraft:strut:mass'
        MASS_COEFFICIENT = 'aircraft:strut:mass_coefficient'
        THICKNESS_TO_CHORD = 'aircraft:strut:thickness_to_chord'

    class TailBoom:
        LENGTH = 'aircraft:tail_boom:length'

    class VerticalTail:
        AREA = 'aircraft:vertical_tail:area'
        ASPECT_RATIO = 'aircraft:vertical_tail:aspect_ratio'
        AVERAGE_CHORD = 'aircraft:vertical_tail:average_chord'
        CHARACTERISTIC_LENGTH = 'aircraft:vertical_tail:characteristic_length'
        FINENESS = 'aircraft:vertical_tail:fineness'
        FORM_FACTOR = 'aircraft:vertical_tail:form_factor'
        LAMINAR_FLOW_LOWER = 'aircraft:vertical_tail:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:vertical_tail:laminar_flow_upper'
        MASS = 'aircraft:vertical_tail:mass'
        MASS_COEFFICIENT = 'aircraft:vertical_tail:mass_coefficient'
        MASS_SCALER = 'aircraft:vertical_tail:mass_scaler'
        MOMENT_ARM = 'aircraft:vertical_tail:moment_arm'
        MOMENT_RATIO = 'aircraft:vertical_tail:moment_ratio'
        NUM_TAILS = 'aircraft:vertical_tail:num_tails'
        ROOT_CHORD = 'aircraft:vertical_tail:root_chord'
        SPAN = 'aircraft:vertical_tail:span'
        SWEEP = 'aircraft:vertical_tail:sweep'
        TAPER_RATIO = 'aircraft:vertical_tail:taper_ratio'
        THICKNESS_TO_CHORD = 'aircraft:vertical_tail:thickness_to_chord'
        VOLUME_COEFFICIENT = 'aircraft:vertical_tail:volume_coefficient'
        WETTED_AREA = 'aircraft:vertical_tail:wetted_area'
        WETTED_AREA_SCALER = 'aircraft:vertical_tail:wetted_area_scaler'

    class Wing:
        AEROELASTIC_TAILORING_FACTOR = \
            'aircraft:wing:aeroelastic_tailoring_factor'

        AIRFOIL_TECHNOLOGY = 'aircraft:wing:airfoil_technology'
        AREA = 'aircraft:wing:area'
        ASPECT_RATIO = 'aircraft:wing:aspect_ratio'
        ASPECT_RATIO_REF = 'aircraft:wing:aspect_ratio_reference'
        AVERAGE_CHORD = 'aircraft:wing:average_chord'
        BENDING_FACTOR = 'aircraft:wing:bending_factor'
        BENDING_MASS = 'aircraft:wing:bending_mass'
        BENDING_MASS_NO_INERTIA = 'aircraft:wing:bending_mass_no_inertia'
        BENDING_MASS_SCALER = 'aircraft:wing:bending_mass_scaler'
        BWB_AFTBODY_MASS = 'aircraft:wing:bwb_aft_body_mass'
        BWB_AFTBODY_MASS_SCALER = 'aircraft:wing:bwb_aft_body_mass_scaler'
        CENTER_CHORD = 'aircraft:wing:center_chord'
        CENTER_DISTANCE = 'aircraft:wing:center_distance'
        CHARACTERISTIC_LENGTH = 'aircraft:wing:characteristic_length'
        CHOOSE_FOLD_LOCATION = 'aircraft:wing:choose_fold_location'
        CHORD_PER_SEMISPAN_DIST = 'aircraft:wing:chord_per_semispan'
        COMPOSITE_FRACTION = 'aircraft:wing:composite_fraction'
        CONTROL_SURFACE_AREA = 'aircraft:wing:control_surface_area'
        CONTROL_SURFACE_AREA_RATIO = 'aircraft:wing:control_surface_area_ratio'
        DETAILED_WING = 'aircraft:wing:detailed_wing'
        DIHEDRAL = 'aircraft:wing:dihedral'
        ENG_POD_INERTIA_FACTOR = 'aircraft:wing:eng_pod_inertia_factor'
        FINENESS = 'aircraft:wing:fineness'
        FLAP_CHORD_RATIO = 'aircraft:wing:flap_chord_ratio'
        FLAP_DEFLECTION_LANDING = 'aircraft:wing:flap_deflection_landing'
        FLAP_DEFLECTION_TAKEOFF = 'aircraft:wing:flap_deflection_takeoff'
        FLAP_DRAG_INCREMENT_OPTIMUM = 'aircraft:wing:flap_drag_increment_optimum'
        FLAP_LIFT_INCREMENT_OPTIMUM = 'aircraft:wing:flap_lift_increment_optimum'
        FLAP_SPAN_RATIO = 'aircraft:wing:flap_span_ratio'
        FLAP_TYPE = 'aircraft:wing:flap_type'
        FOLD_DIMENSIONAL_LOCATION_SPECIFIED = 'aircraft:wing:fold_dimensional_location_specified'
        FOLD_MASS = 'aircraft:wing:fold_mass'
        FOLD_MASS_COEFFICIENT = 'aircraft:wing:fold_mass_coefficient'
        FOLDED_SPAN = 'aircraft:wing:folded_span'
        FOLDED_SPAN_DIMENSIONLESS = 'aircraft:wing:folded_span_dimensionless'
        FOLDING_AREA = 'aircraft:wing:folding_area'
        FORM_FACTOR = 'aircraft:wing:form_factor'
        FUSELAGE_INTERFERENCE_FACTOR = 'aircraft:wing:fuselage_interference_factor'
        GLOVE_AND_BAT = 'aircraft:wing:glove_and_bat'
        HAS_FOLD = 'aircraft:wing:has_fold'
        HAS_STRUT = 'aircraft:wing:has_strut'
        HEIGHT = 'aircraft:wing:height'
        HIGH_LIFT_MASS = 'aircraft:wing:high_lift_mass'
        HIGH_LIFT_MASS_COEFFICIENT = 'aircraft:wing:high_lift_mass_coefficient'
        INCIDENCE = 'aircraft:wing:incidence'
        INPUT_STATION_DIST = 'aircraft:wing:input_station_dist'
        LAMINAR_FLOW_LOWER = 'aircraft:wing:laminar_flow_lower'
        LAMINAR_FLOW_UPPER = 'aircraft:wing:laminar_flow_upper'
        LEADING_EDGE_SWEEP = 'aircraft:wing:leading_edge_sweep'
        LOAD_DISTRIBUTION_CONTROL = 'aircraft:wing:load_distribution_control'
        LOAD_FRACTION = 'aircraft:wing:load_fraction'
        LOAD_PATH_SWEEP_DIST = 'aircraft:wing:load_path_sweep_dist'
        LOADING = 'aircraft:wing:loading'
        LOADING_ABOVE_20 = 'aircraft:wing:loading_above_20'
        MASS = 'aircraft:wing:mass'
        MASS_COEFFICIENT = 'aircraft:wing:mass_coefficient'
        MASS_SCALER = 'aircraft:wing:mass_scaler'
        MATERIAL_FACTOR = 'aircraft:wing:material_factor'
        MAX_CAMBER_AT_70_SEMISPAN = 'aircraft:wing:max_camber_at_70_semispan'
        MAX_LIFT_REF = 'aircraft:wing:max_lift_ref'
        MAX_SLAT_DEFLECTION_LANDING = 'aircraft:wing:max_slat_deflection_landing'
        MAX_SLAT_DEFLECTION_TAKEOFF = 'aircraft:wing:max_slat_deflection_takeoff'
        MAX_THICKNESS_LOCATION = 'aircraft:wing:max_thickness_location'
        MIN_PRESSURE_LOCATION = 'aircraft:wing:min_pressure_location'
        MISC_MASS = 'aircraft:wing:misc_mass'
        MISC_MASS_SCALER = 'aircraft:wing:misc_mass_scaler'
        MOUNTING_TYPE = 'aircraft:wing:mounting_type'
        NUM_FLAP_SEGMENTS = 'aircraft:wing:num_flap_segments'
        NUM_INTEGRATION_STATIONS = 'aircraft:wing:num_integration_stations'
        OPTIMUM_FLAP_DEFLECTION = 'aircraft:wing:optimum_flap_deflection'
        OPTIMUM_SLAT_DEFLECTION = 'aircraft:wing:optimum_slat_deflection'
        ROOT_CHORD = 'aircraft:wing:root_chord'
        SHEAR_CONTROL_MASS = 'aircraft:wing:shear_control_mass'

        SHEAR_CONTROL_MASS_SCALER = \
            'aircraft:wing:shear_control_mass_scaler'

        SLAT_CHORD_RATIO = 'aircraft:wing:slat_chord_ratio'
        SLAT_LIFT_INCREMENT_OPTIMUM = 'aircraft:wing:slat_lift_increment_optimum'
        SLAT_SPAN_RATIO = 'aircraft:wing:slat_span_ratio'
        SPAN = 'aircraft:wing:span'
        SPAN_EFFICIENCY_FACTOR = 'aircraft:wing:span_efficiency_factor'
        SPAN_EFFICIENCY_REDUCTION = 'aircraft:wing:span_efficiency_reduction'
        STRUT_BRACING_FACTOR = 'aircraft:wing:strut_bracing_factor'
        SURFACE_CONTROL_MASS = 'aircraft:wing:surface_ctrl_mass'
        SURFACE_CONTROL_MASS_COEFFICIENT = 'aircraft:wing:surface_ctrl_mass_coefficient'

        SURFACE_CONTROL_MASS_SCALER = \
            'aircraft:wing:surface_ctrl_mass_scaler'

        SWEEP = 'aircraft:wing:sweep'
        TAPER_RATIO = 'aircraft:wing:taper_ratio'
        THICKNESS_TO_CHORD = 'aircraft:wing:thickness_to_chord'
        THICKNESS_TO_CHORD_DIST = 'aircraft:wing:thickness_to_chord_dist'
        THICKNESS_TO_CHORD_REF = 'aircraft:wing:thickness_to_chord_reference'
        THICKNESS_TO_CHORD_ROOT = 'aircraft:wing:thickness_to_chord_root'
        THICKNESS_TO_CHORD_TIP = 'aircraft:wing:thickness_to_chord_tip'
        THICKNESS_TO_CHORD_UNWEIGHTED = 'aircraft:wing:thickness_to_chord_unweighted'
        ULTIMATE_LOAD_FACTOR = 'aircraft:wing:ultimate_load_factor'
        VAR_SWEEP_MASS_PENALTY = 'aircraft:wing:var_sweep_mass_penalty'
        WETTED_AREA = 'aircraft:wing:wetted_area'
        WETTED_AREA_SCALER = 'aircraft:wing:wetted_area_scaler'
        ZERO_LIFT_ANGLE = 'aircraft:wing:zero_lift_angle'


class Dynamic:

    class Mission:
        # all time-dependent variables used during mission analysis
        ALTITUDE = 'altitude'
        ALTITUDE_RATE = 'altitude_rate'
        ALTITUDE_RATE_MAX = 'altitude_rate_max'
        DENSITY = 'density'
        DISTANCE = 'distance'
        DISTANCE_RATE = 'distance_rate'
        DRAG = 'drag'
        DYNAMIC_PRESSURE = 'dynamic_pressure'
        ELECTRIC_POWER = 'electric_power'
        ELECTRIC_POWER_TOTAL = 'electric_power_total'
        # EXIT_AREA = 'exit_area'
        FLIGHT_PATH_ANGLE = 'flight_path_angle'
        FLIGHT_PATH_ANGLE_RATE = 'flight_path_angle_rate'
        FUEL_FLOW_RATE = 'fuel_flow_rate'
        FUEL_FLOW_RATE_NEGATIVE = 'fuel_flow_rate_negative'
        FUEL_FLOW_RATE_NEGATIVE_TOTAL = 'fuel_flow_rate_negative_total'
        FUEL_FLOW_RATE_TOTAL = 'fuel_flow_rate_total'
        HYBRID_THROTTLE = 'hybrid_throttle'
        LIFT = 'lift'
        MACH = 'mach'
        MACH_RATE = 'mach_rate'
        MASS = 'mass'
        MASS_RATE = 'mass_rate'
        NOX_RATE = 'nox_rate'
        NOX_RATE_TOTAL = 'nox_rate_total'
        SPECIFIC_ENERGY = 'specific_energy'
        SPECIFIC_ENERGY_RATE = 'specific_energy_rate'
        SPECIFIC_ENERGY_RATE_EXCESS = 'specific_energy_rate_excess'
        SPEED_OF_SOUND = 'speed_of_sound'
        STATIC_PRESSURE = 'static_pressure'
        TEMPERATURE = 'temperature'
        TEMPERATURE_ENGINE_T4 = 't4'
        THROTTLE = 'throttle'
        THRUST = 'thrust_net'
        THRUST_MAX = 'thrust_net_max'
        THRUST_MAX_TOTAL = 'thrust_net_max_total'
        THRUST_TOTAL = 'thrust_net_total'
        VELOCITY = 'velocity'
        VELOCITY_RATE = 'velocity_rate'


class Mission:

    class Constraints:
        # these can be residuals (for equality constraints),
        # upper bounds, or lower bounds
        MASS_RESIDUAL = 'mission:constraints:mass_residual'
        MAX_MACH = 'mission:constraints:max_mach'
        RANGE_RESIDUAL = 'mission:constraints:range_residual'

    class Design:
        # These values MAY change in design mission, but in off-design
        # they cannot change. In a design mission these are either user inputs
        # or calculated outputs, in off-design they are strictly inputs
        # and do not change.
        CRUISE_ALTITUDE = 'mission:design:cruise_altitude'
        CRUISE_RANGE = 'mission:design:cruise_range'
        FUEL_MASS = 'mission:design:fuel_mass'
        FUEL_MASS_REQUIRED = 'mission:design:fuel_mass_required'
        GROSS_MASS = 'mission:design:gross_mass'
        LIFT_COEFFICIENT = 'mission:design:lift_coefficient'
        LIFT_COEFFICIENT_MAX_FLAPS_UP = 'mission:design:lift_coefficient_max_flaps_up'
        MACH = 'mission:design:mach'
        RANGE = 'mission:design:range'
        RATE_OF_CLIMB_AT_TOP_OF_CLIMB = 'mission:design:rate_of_climb_at_top_of_climb'
        RESERVE_FUEL = 'mission:design:reserve_fuel'
        THRUST_TAKEOFF_PER_ENG = 'mission:design:thrust_takeoff_per_eng'

    class Landing:
        # These are values which have to do with landing
        AIRPORT_ALTITUDE = 'mission:landing:airport_altitude'
        BRAKING_DELAY = 'mission:landing:braking_delay'
        BRAKING_FRICTION_COEFFICIENT = 'mission:landing:braking_friction_coefficient'
        DRAG_COEFFICIENT = 'mission:landing:drag_coefficient'

        DRAG_COEFFICIENT_FLAP_INCREMENT = \
            'mission:landing:drag_coefficient_flap_increment'

        FIELD_LENGTH = 'mission:landing:field_length'
        FLARE_RATE = 'mission:landing:flare_rate'
        GLIDE_TO_STALL_RATIO = 'mission:landing:glide_to_stall_ratio'
        GROUND_DISTANCE = 'mission:landing:ground_distance'
        INITIAL_ALTITUDE = 'mission:landing:initial_altitude'
        INITIAL_MACH = 'mission:landing:initial_mach'
        INITIAL_VELOCITY = 'mission:landing:initial_velocity'

        LIFT_COEFFICIENT_FLAP_INCREMENT = \
            'mission:landing:lift_coefficient_flap_increment'

        LIFT_COEFFICIENT_MAX = 'mission:landing:lift_coefficient_max'
        MAXIMUM_FLARE_LOAD_FACTOR = 'mission:landing:maximum_flare_load_factor'
        MAXIMUM_SINK_RATE = 'mission:landing:maximum_sink_rate'
        OBSTACLE_HEIGHT = 'mission:landing:obstacle_height'
        ROLLING_FRICTION_COEFFICIENT = 'mission:landing:rolling_friction_coefficient'
        SPOILER_DRAG_COEFFICIENT = 'mission:landing:spoiler_drag_coefficient'
        SPOILER_LIFT_COEFFICIENT = 'mission:landing:spoiler_lift_coefficient'
        STALL_VELOCITY = 'mission:landing:stall_velocity'
        TOUCHDOWN_MASS = 'mission:landing:touchdown_mass'
        TOUCHDOWN_SINK_RATE = 'mission:landing:touchdown_sink_rate'

    class Objectives:
        # these are values that can be fed to the optimizer as objectives,
        # they may often be composite and/or regularized
        FUEL = 'mission:objectives:fuel'
        RANGE = 'mission:objectives:range'

    class Summary:
        # These values are inputs and outputs to/from dynamic analysis
        # for the given mission (whether it is design or off-design).
        # In on-design these may be constrained to design values, but
        # in off-design they independently represent the final analysis
        # based on the user-selection.
        CRUISE_MACH = 'mission:summary:cruise_mach'
        CRUISE_MASS_FINAL = 'mission:summary:cruise_mass_final'
        FUEL_FLOW_SCALER = 'mission:summary:fuel_flow_scaler'
        GROSS_MASS = 'mission:summary:gross_mass'
        RANGE = 'mission:summary:range'
        TOTAL_FUEL_MASS = 'mission:summary:total_fuel_mass'

    class Takeoff:
        # These are values which have to do with takeoff
        AIRPORT_ALTITUDE = 'mission:takeoff:airport_altitude'
        ANGLE_OF_ATTACK_RUNWAY = 'mission:takeoff:angle_of_attack_runway'
        ASCENT_DURATION = 'mission:takeoff:ascent_duration'
        ASCENT_T_INTIIAL = 'mission:takeoff:ascent_t_initial'
        BRAKING_FRICTION_COEFFICIENT = 'mission:takeoff:braking_friction_coefficient'
        DECISION_SPEED_INCREMENT = 'mission:takeoff:decision_speed_increment'

        DRAG_COEFFICIENT_FLAP_INCREMENT = \
            'mission:takeoff:drag_coefficient_flap_increment'

        DRAG_COEFFICIENT_MIN = 'mission:takeoff:drag_coefficient_min'
        FIELD_LENGTH = 'mission:takeoff:field_length'
        FINAL_ALTITUDE = 'mission:takeoff:final_altitude'
        FINAL_MACH = 'mission:takeoff:final_mach'
        FINAL_MASS = 'mission:takeoff:final_mass'
        FINAL_VELOCITY = 'mission:takeoff:final_velocity'
        FUEL_SIMPLE = 'mission:takeoff:fuel_simple'
        GROUND_DISTANCE = 'mission:takeoff:ground_distance'

        LIFT_COEFFICIENT_FLAP_INCREMENT = \
            'mission:takeoff:lift_coefficient_flap_increment'

        LIFT_COEFFICIENT_MAX = 'mission:takeoff:lift_coefficient_max'
        LIFT_OVER_DRAG = 'mission:takeoff:lift_over_drag'
        OBSTACLE_HEIGHT = 'mission:takeoff:obstacle_height'
        ROLLING_FRICTION_COEFFICIENT = 'mission:takeoff:rolling_friction_coefficient'
        ROTATION_SPEED_INCREMENT = 'mission:takeoff:rotation_speed_increment'
        ROTATION_VELOCITY = 'mission:takeoff:rotation_velocity'
        SPOILER_DRAG_COEFFICIENT = 'mission:takeoff:spoiler_drag_coefficient'
        SPOILER_LIFT_COEFFICIENT = 'mission:takeoff:spoiler_lift_coefficient'
        THRUST_INCIDENCE = 'mission:takeoff:thrust_incidence'

    class Taxi:
        DURATION = 'mission:taxi:duration'
        MACH = 'mission:taxi:mach'


class Settings:
    EQUATIONS_OF_MOTION = 'settings:equations_of_motion'
    MASS_METHOD = 'settings:mass_method'
    VERBOSITY = 'settings:verbosity'



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.mission.gasp_based.ode.unsteady_solved.unsteady_solved_eom import \
    UnsteadySolvedEOM
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class TestUnsteadyFlightEOM(unittest.TestCase):

    def _test_unsteady_flight_eom(self, ground_roll=False):
        nn = 5

        p = om.Problem()
        p.model.add_subsystem("eom",
                              UnsteadySolvedEOM(num_nodes=nn, ground_roll=ground_roll),
                              promotes_inputs=["*"],
                              promotes_outputs=["*"])

        p.setup(force_alloc_complex=True)

        p.set_val("TAS", 250, units="kn")
        p.set_val(Dynamic.Mission.MASS, 175_000, units="lbm")
        p.set_val(Dynamic.Mission.THRUST_TOTAL, 20_000, units="lbf")
        p.set_val(Dynamic.Mission.LIFT, 175_000, units="lbf")
        p.set_val(Dynamic.Mission.DRAG, 20_000, units="lbf")
        p.set_val(Aircraft.Wing.INCIDENCE, 0.0, units="deg")

        if not ground_roll:
            p.set_val("alpha", 0.0, units="deg")
            p.set_val(Dynamic.Mission.FLIGHT_PATH_ANGLE, 0, units="deg")
            p.set_val("dh_dr", 0, units=None)
            p.set_val("d2h_dr2", 0, units="1/m")

        p.run_model()

        # p.model.list_inputs()
        # p.model.list_outputs(print_arrays=True, units=True)

        dt_dr = p.get_val("dt_dr", units="h/NM")
        normal_force = p.get_val("normal_force", units="lbf")
        load_factor = p.get_val("load_factor", units="unitless")
        fuselage_pitch = p.get_val("fuselage_pitch", units="deg")

        # True airspeed in level flight is dr_dt
        assert_near_equal(1/dt_dr, 250.0 * np.ones(nn), tolerance=1.0E-12)

        # Normal force in balanced level flight is 0.0
        assert_near_equal(normal_force, np.zeros(nn), tolerance=1.0E-12)

        # Fuselage pitch balanced level flight with zero alpha and wing incidence is 0.0
        assert_near_equal(fuselage_pitch, np.zeros(nn), tolerance=1.0E-12)

        # Load factor balanced level flight with zero alpha and wing incidence is 1.0
        assert_near_equal(load_factor, np.ones(nn), tolerance=1.0E-12)

        if not ground_roll:
            dgam_dt = p.get_val("dgam_dt", units="deg/s")
            dgam_dt_approx = p.get_val("dgam_dt_approx", units="deg/s")

            # Both approximate and computed dgam_dt should be zero.
            assert_near_equal(dgam_dt, np.zeros(nn), tolerance=1.0E-12)
            assert_near_equal(dgam_dt_approx, np.zeros(nn), tolerance=1.0E-12)

        p.set_val("TAS", 250 + 10 * np.random.rand(nn), units="kn")
        p.set_val(Dynamic.Mission.MASS, 175_000 + 1000 * np.random.rand(nn), units="lbm")
        p.set_val(Dynamic.Mission.THRUST_TOTAL, 20_000 +
                  100 * np.random.rand(nn), units="lbf")
        p.set_val(Dynamic.Mission.LIFT, 175_000 + 1000 * np.random.rand(nn), units="lbf")
        p.set_val(Dynamic.Mission.DRAG, 20_000 + 100 * np.random.rand(nn), units="lbf")
        p.set_val(Aircraft.Wing.INCIDENCE, np.random.rand(1), units="deg")

        if not ground_roll:
            p.set_val("alpha", 5 * np.random.rand(nn), units="deg")
            p.set_val(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                      5 * np.random.rand(nn), units="deg")
            p.set_val("dh_dr", 0.1 * np.random.rand(nn), units=None)
            p.set_val("d2h_dr2", 0.01 * np.random.rand(nn), units="1/m")

        p.run_model()

        with np.printoptions(linewidth=1024):
            cpd = p.check_partials(method="cs")
        assert_check_partials(cpd)

    def test_unsteady_flight_eom(self):
        for ground_roll in True, False:
            with self.subTest(msg=f"ground_roll={ground_roll}"):
                self._test_unsteady_flight_eom(ground_roll=ground_roll)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from dymos.models.atmosphere import USatm1976Comp
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from parameterized import parameterized

from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.subsystems.premission import CorePreMission
from aviary.interface.default_phase_info.height_energy import aero, prop, geom
from aviary.utils.aviary_values import AviaryValues, get_items
from aviary.utils.functions import set_aviary_initial_values
from aviary.utils.named_values import NamedValues
from aviary.validation_cases.validation_tests import (get_flops_inputs,
                                                      get_flops_outputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.variable_info.variables_in import VariablesIn
from aviary.variable_info.enums import LegacyCode

FLOPS = LegacyCode.FLOPS
GASP = LegacyCode.GASP


class TabularAeroGroupFileTest(unittest.TestCase):
    # Test drag comp with data from file, structured grid
    def setUp(self):
        self.prob = om.Problem()
        aviary_options = AviaryValues()

        CDI_table = "subsystems/aerodynamics/flops_based/test/large_single_aisle_1_CDI_polar.csv"
        CD0_table = "subsystems/aerodynamics/flops_based/test/large_single_aisle_1_CD0_polar.csv"

        kwargs = {'method': 'tabular', 'CDI_data': CDI_table,
                  'CD0_data': CD0_table}

        aero_builder = CoreAerodynamicsBuilder(code_origin=FLOPS)

        self.prob.model.add_subsystem(
            "aero",
            aero_builder.build_mission(
                num_nodes=1, aviary_inputs=aviary_options, **kwargs),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case(self):
        # TODO currently no way to use FLOPS test case data for mission components
        # test data from large_single_aisle_2 climb profile
        # tabular aero was set to large_single_aisle_1, expected value adjusted accordingly
        self.prob.set_val(
            Dynamic.Mission.VELOCITY,
            val=115,
            units='m/s')  # convert from knots to ft/s
        self.prob.set_val(Dynamic.Mission.ALTITUDE, val=10582, units='m')
        self.prob.set_val(Dynamic.Mission.MASS, val=80442, units='kg')
        self.prob.set_val(Dynamic.Mission.MACH, val=0.3876, units='unitless')
        # 1344.5? 'reference' vs 'calculated'?
        self.prob.set_val(Aircraft.Wing.AREA, val=1341, units='ft**2')
        # calculated from online atmospheric table
        self.prob.set_val(Dynamic.Mission.DENSITY, val=0.88821, units='kg/m**3')

        self.prob.run_model()

        # We know that computed drag (from FLOPS) is higher than what this tabular data
        # computes. Use loose tolerance.
        tol = .03

        assert_near_equal(
            self.prob.get_val(Dynamic.Mission.DRAG, units='N'), 53934.78861492, tol
        )  # check the value of each output

        # TODO resolve partials wrt gravity (decide on implementation of gravity)
        partial_data = self.prob.check_partials(
            out_stream=None, method="cs", step=1.1e-40)
        assert_check_partials(
            partial_data, atol=1e-9, rtol=1e-12
        )  # check the partial derivatives


class TabularAeroGroupDataTest(unittest.TestCase):
    # Test tabular drag comp with training data, structured grid
    def setUp(self):
        self.prob = om.Problem()
        aviary_options = AviaryValues()

        CDI_table = _default_CDI_data()
        CDI_values = CDI_table.get_val('lift_dependent_drag_coefficient')
        # CDI_table.delete('lift_dependent_drag_coefficient')
        CD0_table = _default_CD0_data()
        CD0_values = CD0_table.get_val('zero_lift_drag_coefficient')
        # CD0_table.delete('zero_lift_drag_coefficient')

        drag_data = om.ExecComp()
        drag_data.add_output('zero_lift_drag_coefficient_train',
                             CD0_values, units='unitless')
        drag_data.add_output('lift_dependent_drag_coefficient_train',
                             CDI_values, units='unitless')

        self.prob.model.add_subsystem(
            "drag_data",
            drag_data,
            promotes_outputs=['*']
        )

        kwargs = {'method': 'tabular', 'CDI_data': CDI_table, 'CD0_data': CD0_table, }
        #   'training_data': True}

        aero_builder = CoreAerodynamicsBuilder(code_origin=FLOPS)

        self.prob.model.add_subsystem(
            "aero",
            aero_builder.build_mission(
                num_nodes=1, aviary_inputs=aviary_options, **kwargs),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case(self):
        # TODO currently no way to use FLOPS test case data for mission components
        # test data from large_single_aisle_2 climb profile
        # tabular aero was set to large_single_aisle_1 data, expected value adjusted accordingly
        self.prob.set_val(
            Dynamic.Mission.VELOCITY,
            val=115,
            units='m/s')  # convert from knots to ft/s
        self.prob.set_val(Dynamic.Mission.ALTITUDE, val=10582, units='m')
        self.prob.set_val(Dynamic.Mission.MASS, val=80442, units='kg')
        self.prob.set_val(Dynamic.Mission.MACH, val=0.3876, units='unitless')
        # 1344.5? 'reference' vs 'calculated'?
        self.prob.set_val(Aircraft.Wing.AREA, val=1341, units='ft**2')
        # calculated from online atmospheric table
        self.prob.set_val(Dynamic.Mission.DENSITY, val=0.88821, units='kg/m**3')

        self.prob.run_model()

        # We know that computed drag (from FLOPS) is higher than what this tabular data
        # computes. Use loose tolerance.
        tol = .03

        assert_near_equal(
            self.prob.get_val(Dynamic.Mission.DRAG, units='N'), 53934.78861492, tol
        )  # check the value of each output

        # TODO resolve partials wrt gravity (decide on implementation of gravity)
        partial_data = self.prob.check_partials(
            out_stream=None, method="cs", step=1.1e-40)
        assert_check_partials(
            partial_data, atol=1e-9, rtol=1e-12
        )  # check the partial derivatives


data_sets = ['LargeSingleAisle1FLOPS', 'LargeSingleAisle2FLOPS', 'N3CC']


class ComputedVsTabularTest(unittest.TestCase):

    @parameterized.expand(data_sets, name_func=print_case)
    def test_case(self, case_name):
        flops_inputs = get_flops_inputs(case_name)
        flops_outputs = get_flops_outputs(case_name)

        flops_inputs.set_val(Aircraft.Design.LIFT_DEPENDENT_DRAG_COEFF_FACTOR, 0.9)
        flops_inputs.set_val(Aircraft.Design.ZERO_LIFT_DRAG_COEFF_FACTOR, 1.1)
        flops_inputs.set_val(Aircraft.Design.SUBSONIC_DRAG_COEFF_FACTOR, 0.9)
        flops_inputs.set_val(Aircraft.Design.SUPERSONIC_DRAG_COEFF_FACTOR, 1.1)

        key = Aircraft.Propulsion.TOTAL_NUM_ENGINES
        flops_inputs.set_val(key, flops_outputs.get_val(key))

        key = Aircraft.Propulsion.TOTAL_SCALED_SLS_THRUST
        flops_inputs.set_val(key, *(flops_outputs.get_item(key)))

        mass, units = flops_inputs.get_item(Mission.Design.GROSS_MASS)
        mass = mass * 0.92

        alt = 10582
        alt_units = 'm'

        vel = 115
        vel_units = 'm/s'

        dynamic_inputs = AviaryValues()

        dynamic_inputs.set_val(Dynamic.Mission.VELOCITY, val=vel, units=vel_units)
        dynamic_inputs.set_val(Dynamic.Mission.ALTITUDE, val=alt, units=alt_units)
        dynamic_inputs.set_val(Dynamic.Mission.MASS, val=mass, units=units)

        prob = _get_computed_aero_data_at_altitude(alt, alt_units)

        sos = prob.get_val(Dynamic.Mission.SPEED_OF_SOUND, vel_units)
        mach = vel / sos

        dynamic_inputs.set_val(Dynamic.Mission.MACH, val=mach, units='unitless')

        key = Dynamic.Mission.DENSITY
        units = 'kg/m**3'
        val = prob.get_val(key, units)

        dynamic_inputs.set_val(key, val=val, units=units)

        key = Dynamic.Mission.TEMPERATURE
        units = 'degR'
        val = prob.get_val(key, units)

        dynamic_inputs.set_val(key, val=val, units=units)

        key = Dynamic.Mission.STATIC_PRESSURE
        units = 'N/m**2'
        val = prob.get_val(key, units)

        dynamic_inputs.set_val(key, val=val, units=units)

        prob = _run_computed_aero_harness(flops_inputs, dynamic_inputs, 1)

        computed_drag = prob.get_val(Dynamic.Mission.DRAG, 'N')

        CDI_data, CD0_data = _computed_aero_drag_data(
            flops_inputs, *_design_altitudes.get_item(case_name))

        prob = om.Problem()

        kwargs = {'method': 'tabular', 'CDI_data': CDI_data,
                  'CD0_data': CD0_data}

        aero_builder = CoreAerodynamicsBuilder(code_origin=FLOPS)

        prob.model.add_subsystem(
            "aero",
            aero_builder.build_mission(
                num_nodes=1, aviary_inputs=flops_inputs, **kwargs),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        for (key, (val, units)) in dynamic_inputs:
            try:
                prob.set_val(key, val, units)

            except:
                pass  # unused variable

        for (key, (val, units)) in get_items(flops_inputs):
            try:
                prob.set_val(key, val, units)

            except:
                # Should be an option or an overridden output.
                continue

        prob.run_model()

        tabular_drag = prob.get_val(Dynamic.Mission.DRAG, 'N')

        assert_near_equal(tabular_drag, computed_drag, 0.005)


# region - hardcoded data from large_single_aisle_1 (fixed alt cruise variant) FLOPS model
def _default_CD0_data():
    alt_range = _make_default_alt_range()  # len == 13
    mach_range = _make_default_mach_range()  # len == 12

    CD0 = [
        [
            0.02182, 0.02228, 0.02278, 0.02332,
            0.02389, 0.02451, 0.02518, 0.0259,
            0.02689, 0.02801, 0.02921, 0.03049],
        [
            0.03184, 0.02043, 0.02086, 0.02132,
            0.0218, 0.02233, 0.02289, 0.02349,
            0.02414, 0.02504, 0.02605, 0.02713],
        [
            0.02827, 0.02949, 0.01946, 0.01987,
            0.02029, 0.02075, 0.02123, 0.02176,
            0.02232, 0.02293, 0.02377, 0.02471],
        [
            0.0257, 0.02676, 0.02788, 0.0187,
            0.01908, 0.01949, 0.01992, 0.02038,
            0.02088, 0.02141, 0.02199, 0.02278],
        [
            0.02366, 0.0246, 0.0256, 0.02665,
            0.01824, 0.0186, 0.01899, 0.0194,
            0.01984, 0.02031, 0.02082, 0.02137],
        [
            0.02212, 0.02297, 0.02386, 0.02481,
            0.02581, 0.01828, 0.01863, 0.019,
            0.0194, 0.01982, 0.02027, 0.02076],
        [
            0.02129, 0.02201, 0.02282, 0.02368,
            0.02458, 0.02554, 0.01843, 0.01878,
            0.01914, 0.01953, 0.01994, 0.02039],
        [
            0.02087, 0.02139, 0.02209, 0.02289,
            0.02373, 0.02461, 0.02555, 0.01864,
            0.01898, 0.01934, 0.01972, 0.02014],
        [
            0.02058, 0.02105, 0.02157, 0.02227,
            0.02305, 0.02388, 0.02476, 0.02569,
            0.01909, 0.01943, 0.01979, 0.02017],
        [
            0.02058, 0.02101, 0.02149, 0.02199,
            0.02269, 0.02347, 0.02429, 0.02516,
            0.02608, 0.02021, 0.02054, 0.02089],
        [
            0.02127, 0.02168, 0.02211, 0.02258,
            0.02308, 0.02377, 0.02454, 0.02535,
            0.02622, 0.02713, 0.02222, 0.02255],
        [
            0.0229, 0.02328, 0.02368, 0.02411,
            0.02457, 0.02507, 0.02575, 0.02651,
            0.02732, 0.02818, 0.02908, 0.03704],
        [
            0.03737, 0.03772, 0.03809, 0.03849,
            0.03891, 0.03937, 0.03987, 0.04054,
            0.0413, 0.0421, 0.04294, 0.04384]]

    # mach_list, alt_list = np.meshgrid(mach_range, alt_range)

    CD0 = np.array(CD0)  # .flatten()
    # mach_list = np.array(mach_list).flatten()
    # alt_list = np.array(alt_list).flatten()

    CD0_data = NamedValues()
    CD0_data.set_val(Dynamic.Mission.ALTITUDE, alt_range, 'ft')
    CD0_data.set_val(Dynamic.Mission.MACH, mach_range)
    CD0_data.set_val('zero_lift_drag_coefficient', CD0)

    return CD0_data


def _default_CDI_data():
    mach_range = _make_default_mach_range()  # len == 12
    cl_range = _make_default_cl_range()  # len == 15

    CDI = [
        [
            0.00114, 0.00138, 0.0019, 0.0027, 0.00378,
            0.00511, 0.00669, 0.00837, 0.0104, 0.01295,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 0
        [
            0.00113, 0.00137, 0.0019, 0.0027, 0.00377,
            0.00511, 0.00669, 0.00837, 0.0104, 0.01295,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 1
        [
            0.00114, 0.00138, 0.0019, 0.0027, 0.00378,
            0.00511, 0.00669, 0.00837, 0.0104, 0.01295,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 2
        [
            0.00117, 0.00138, 0.0019, 0.0027, 0.00379,
            0.00513, 0.0067, 0.00837, 0.01041, 0.01295,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 3
        [
            0.00121, 0.0014, 0.0019, 0.0027, 0.00381,
            0.00515, 0.0067, 0.00838, 0.01042, 0.01294,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 4
        [
            0.00132, 0.00144, 0.0019, 0.00271, 0.00387,
            0.00522, 0.00673, 0.00839, 0.01043, 0.01294,
            0.01558, 0.01837, 0.0215, 0.025, 0.02888],  # 5
        [
            0.00142, 0.00148, 0.00192, 0.00273, 0.00392,
            0.00529, 0.00675, 0.00841, 0.01045, 0.01293,
            0.01556, 0.01841, 0.02169, 0.02542, 0.02961],  # 6
        [
            0.0015, 0.00152, 0.00194, 0.00276, 0.00397,
            0.00534, 0.00678, 0.00842, 0.01046, 0.0129,
            0.01562, 0.01891, 0.02291, 0.02769, 0.03327],  # 7
        [
            0.00166, 0.00159, 0.00196, 0.00279, 0.00406,
            0.00549, 0.00693, 0.00855, 0.01065, 0.01328,
            0.01686, 0.02156, 0.02667, 0.03227, 0.03832],  # 8
        [
            0.00208, 0.00175, 0.00201, 0.00285, 0.00428,
            0.00592, 0.0076, 0.00965, 0.01245, 0.01657,
            0.0215, 0.02697, 0.0341, 0.04264, 0.0526],  # 9
        [
            0.00336, 0.00221, 0.00209, 0.003, 0.00494,
            0.00757, 0.01074, 0.01452, 0.01987, 0.02654,
            0.03295, 0.03957, 0.04634, 0.05334, 0.06059],  # 10
        [
            0.00636, 0.00327, 0.00224, 0.00324, 0.00627,
            0.01154, 0.01915, 0.02577, 0.03848, 0.05251,
            0.05737, 0.06445, 0.0703, 0.07582, 0.08127]]  # 11

    # cl_list, mach_list = np.meshgrid(cl_range, mach_range)

    CDI = np.array(CDI)  # .flatten()
    # cl_list = np.array(cl_list).flatten()
    # mach_list = np.array(mach_list).flatten()
    CDI_data = NamedValues()
    CDI_data.set_val(Dynamic.Mission.MACH, mach_range)
    CDI_data.set_val('lift_coefficient', cl_range)
    CDI_data.set_val('lift_dependent_drag_coefficient', CDI)

    return CDI_data


def _make_default_alt_range():
    alt_range = [
        0, 5000, 10000, 15000,
        20000, 25000, 30000, 35000,
        40000, 45000, 50000, 55000,
        60000]

    return alt_range


def _make_default_mach_range():
    mach_range = [
        0.200, 0.300, 0.400, 0.500,
        0.600, 0.700, 0.750, 0.775,
        0.800, 0.825, 0.850, 0.875]

    return mach_range


def _make_default_cl_range():
    cl_range = [
        0.15, 0.2, 0.25, 0.3,
        0.35, 0.4, 0.45, 0.5,
        0.55, 0.6, 0.65, 0.7,
        0.75, 0.8, 0.85]

    return cl_range
# endregion - hardcoded data from large_single_aisle_1 (fixed alt cruise variant) FLOPS model


# region - computed aero data
def _computed_aero_drag_data(flops_inputs: AviaryValues, design_altitude, units):
    nsteps = 12

    start = 0.
    stop = 70000.
    step = (stop - start) / nsteps
    alt = np.arange(start, stop, step)  # ft

    start = 0.2
    stop = 1.
    step = (stop - start) / nsteps
    seed = np.arange(start, stop, step)

    mach = np.repeat(seed, nsteps)  # unitless
    CL = np.tile(seed, nsteps)  # unitless

    S = flops_inputs.get_val(Aircraft.Wing.AREA, 'ft**2')

    # calculate temperature (degR), static pressure (lbf/ft**2), and mass (lbm) at design
    # altitude from lift coefficients and Mach numbers
    prob: om.Problem = _get_computed_aero_data_at_altitude(design_altitude, units)
    T = prob.get_val(Dynamic.Mission.TEMPERATURE, 'degR')
    P = prob.get_val(Dynamic.Mission.STATIC_PRESSURE, 'lbf/ft**2')

    mass = lift = CL * S * 0.5 * 1.4 * P * mach**2  # lbf -> lbm * 1g

    # calculate lift-dependent drag coefficient table, including pressure drag effects
    nn = len(mass)

    dynamic_inputs = AviaryValues()

    dynamic_inputs.set_val(Dynamic.Mission.MACH, val=mach)
    dynamic_inputs.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P, units='lbf/ft**2')
    dynamic_inputs.set_val(Dynamic.Mission.TEMPERATURE, val=T, units='degR')
    dynamic_inputs.set_val(Dynamic.Mission.MASS, val=mass, units='lbm')

    prob = _run_computed_aero_harness(flops_inputs, dynamic_inputs, nn)

    CDI: np.ndarray = prob.get_val('CDI')
    CDI = np.reshape(CDI.flatten(), (nsteps, nsteps))

    CDI_data = NamedValues()
    CDI_data.set_val(Dynamic.Mission.MACH, seed)
    CDI_data.set_val('lift_coefficient', seed)
    CDI_data.set_val('lift_dependent_drag_coefficient', CDI)

    # calculate lift-independent drag coefficient table
    mass, units = flops_inputs.get_item(Mission.Design.GROSS_MASS)
    mass = mass * 0.9

    mach = seed
    nn = len(mach)

    dynamic_inputs = AviaryValues()

    dynamic_inputs.set_val(Dynamic.Mission.MACH, val=mach)
    dynamic_inputs.set_val(Dynamic.Mission.MASS, val=mass, units=units)

    CD0 = []

    for h in alt:
        prob: om.Problem = _get_computed_aero_data_at_altitude(h, 'ft')
        T = prob.get_val(Dynamic.Mission.TEMPERATURE, 'degR')
        P = prob.get_val(Dynamic.Mission.STATIC_PRESSURE, 'lbf/ft**2')

        dynamic_inputs.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P, units='lbf/ft**2')
        dynamic_inputs.set_val(Dynamic.Mission.TEMPERATURE, val=T, units='degR')

        prob = _run_computed_aero_harness(flops_inputs, dynamic_inputs, nn)

        CD0.append(prob.get_val('CD0'))

    CD0 = np.array(CD0)

    CD0_data = NamedValues()
    CD0_data.set_val(Dynamic.Mission.ALTITUDE, alt, 'ft')
    CD0_data.set_val(Dynamic.Mission.MACH, seed)
    CD0_data.set_val('zero_lift_drag_coefficient', CD0)

    return (CDI_data, CD0_data)


def _get_computed_aero_data_at_altitude(altitude, units):
    prob = om.Problem()

    prob.model.add_subsystem(
        'atm', USatm1976Comp(num_nodes=1),
        promotes_inputs=[('h', Dynamic.Mission.ALTITUDE)],
        promotes_outputs=[
            ('sos', Dynamic.Mission.SPEED_OF_SOUND), ('rho', Dynamic.Mission.DENSITY),
            ('temp', Dynamic.Mission.TEMPERATURE), ('pres', Dynamic.Mission.STATIC_PRESSURE)])

    prob.setup()

    prob.set_val(Dynamic.Mission.ALTITUDE, altitude, units)

    prob.run_model()

    return prob


def _run_computed_aero_harness(flops_inputs, dynamic_inputs, num_nodes):
    prob = om.Problem(
        _ComputedAeroHarness(num_nodes=num_nodes, aviary_options=flops_inputs))

    prob.setup()

    for (key, (val, units)) in get_items(dynamic_inputs):
        try:
            prob.set_val(key, val, units)

        except:
            pass  # unused variable

    for (key, (val, units)) in get_items(flops_inputs):
        try:
            prob.set_val(key, val, units)

        except:
            # Should be an option or an overridden output.
            continue

    prob.run_model()

    return prob


class _ComputedAeroHarness(om.Group):
    '''
    Calculate drag and drag polars.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', types=int)

        options.declare(
            'gamma', default=1.4, desc='Ratio of specific heats for air.')

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        options = self.options

        nn = options['num_nodes']
        gamma = options['gamma']
        aviary_options: AviaryValues = options['aviary_options']

        # replace debug_no_mass flag with list of subsystem builders without mass
        core_subsystems = [prop, geom, aero]

        # Upstream static analysis for aero
        pre_mission: om.Group = self.add_subsystem(
            'pre_mission',
            CorePreMission(aviary_options=aviary_options,
                           subsystems=core_subsystems),
            promotes_inputs=['aircraft:*', 'mission:*'],
            promotes_outputs=['aircraft:*', 'mission:*'])

        kwargs = {'method': 'computed', 'gamma': gamma}

        aero_builder = CoreAerodynamicsBuilder(code_origin=FLOPS)

        self.add_subsystem(
            "aero",
            aero_builder.build_mission(
                num_nodes=nn, aviary_inputs=aviary_options, **kwargs),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        key = Aircraft.Engine.SCALED_SLS_THRUST
        val, units = aviary_options.get_item(key)
        pre_mission.set_input_defaults(key, val, units)

        self.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=aviary_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        set_aviary_initial_values(self, aviary_options)


_design_altitudes = AviaryValues({
    'LargeSingleAisle1FLOPS': (41000, 'ft'), 'LargeSingleAisle2FLOPS': (41000, 'ft'), 'N3CC': (43000, 'ft')})
# endregion - computed aero data


if __name__ == "__main__":
    unittest.main()
    # test = TabularAeroGroupDataTest()
    # test.setUp()
    # test.test_case()



'''
OpenMDAO component for aerostructural analysis using OpenAeroStruct.

This analysis is based on the aircraft_for_bench_FwFm.csv input data representing a
single-aisle commercial transport aircraft.  The user_mesh method is currently
hard-coded with values taken from the aircraft_for_bench_FwFm data, but should be coded
to use the Aircraft.Wing.* variables for a more general capability.

The OAStructures class performs a structural analysis of the given wing
by applying aeroelastic loads computed at the cruise condition and a
2.5g maneuver at Mach 0.64 at sea level.  The optimization determines
the optimum wing skin thickness, spar cap thickness, wing twist, wing t/c
and maneuver angle of attack that satisfies strength constraints.

The only Aviary input driving the design is fuel mass, but other variables
may be included as well.

OAStructures returns the optimized wing mass and the fuel mass burned.
Currently, only the wing mass is used to override the Aviary variable
Aircraft.Wing.MASS.

'''

import time
import warnings

import numpy as np
import openmdao.api as om

try:
    import ambiance
except ImportError:
    raise ImportError(
        "ambiance package not found. You can install it by running 'pip install ambiance'.")

try:
    import openaerostruct
except ImportError:
    raise ImportError(
        "openaerostruct package not found. You can install it by running 'pip install openaerostruct'.")

from ambiance import Atmosphere

from openaerostruct.integration.aerostruct_groups import AerostructGeometry, AerostructPoint
from openaerostruct.structures.wingbox_fuel_vol_delta import WingboxFuelVolDelta


def user_mesh():
    # generate a user defined mesh which is model specific

    # Planform specifications
    half_span = 17.9573
    kink_location = 4.9544

    root_chord = 5.5668
    kink_chord = 4.1302
    tip_chord = 1.5084
    inboard_LE_sweep = 25.0
    outboard_LE_sweep = 25.0

    # Mesh specifications
    nx = 2
    ny_outboard = 5
    ny_inboard = 3

    # Initialize the 3-D mesh object. Indexing: Chordwise, spanwise, then the 3-D coordinates.
    # We use ny_inboard+ny_outboard-1 because the 2 segments share the nodes where they connect.
    mesh = np.zeros((nx, ny_inboard + ny_outboard - 1, 3))

    # The form of this 3-D array can be confusing initially.
    # For each node, we are providing the x, y, and z coordinates.
    # x is streamwise, y is spanwise, and z is up.
    # For example, the node for the leading edge at the tip would be specified as mesh[0, 0, :] = np.array([x, y, z]).
    # And the node at the trailing edge at the root would be mesh[nx-1, ny-1, :] = np.array([x, y, z]).
    # We only provide the right half of the wing here because we use symmetry.
    # Print elements of the mesh to better understand the form.

    ####### THE Z-COORDINATES ######
    # Assume no dihedral, so set the z-coordinate for all the points to 0.
    mesh[:, :, 2] = 0.0

    ####### THE Y-COORDINATES ######
    # Using uniform spacing for the spanwise locations of all the nodes within each of the two trapezoidal segments:
    # Outboard
    mesh[:, :ny_outboard, 1] = np.linspace(half_span, kink_location, ny_outboard)
    # Inboard
    mesh[:, ny_outboard: ny_outboard + ny_inboard,
         1] = np.linspace(kink_location, 0, ny_inboard)[1:]

    ###### THE X-COORDINATES ######
    # Start with the leading edge and create some intermediate arrays that we will use
    x_LE = np.zeros(ny_inboard + ny_outboard - 1)

    array_for_inboard_leading_edge_x_coord = np.linspace(
        0, kink_location, ny_inboard) * np.tan(
        inboard_LE_sweep / 180.0 * np.pi
    )

    array_for_outboard_leading_edge_x_coord = (
        np.linspace(0, half_span - kink_location, ny_outboard) *
        np.tan(outboard_LE_sweep / 180.0 * np.pi)
        + np.ones(ny_outboard) * array_for_inboard_leading_edge_x_coord[-1]
    )

    x_LE[:ny_inboard] = array_for_inboard_leading_edge_x_coord
    x_LE[ny_inboard: ny_inboard + ny_outboard] = array_for_outboard_leading_edge_x_coord[1:]

    # Then the trailing edge
    x_TE = np.zeros(ny_inboard + ny_outboard - 1)

    array_for_inboard_trailing_edge_x_coord = np.linspace(
        array_for_inboard_leading_edge_x_coord[0] + root_chord,
        array_for_inboard_leading_edge_x_coord[-1] + kink_chord,
        ny_inboard,
    )

    array_for_outboard_trailing_edge_x_coord = np.linspace(
        array_for_outboard_leading_edge_x_coord[0] + kink_chord,
        array_for_outboard_leading_edge_x_coord[-1] + tip_chord,
        ny_outboard,
    )

    x_TE[:ny_inboard] = array_for_inboard_trailing_edge_x_coord
    x_TE[ny_inboard: ny_inboard + ny_outboard] = array_for_outboard_trailing_edge_x_coord[1:]

    for i in range(0, ny_inboard + ny_outboard - 1):
        mesh[:, i, 0] = np.linspace(np.flip(x_LE)[i], np.flip(x_TE)[i], nx)

    return mesh


class OAStructures(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('symmetry', default=True,
                             desc='wing symmetry (True or False)')
        self.options.declare('chord_cos_spacing', default=0,
                             desc='chordwise cosine spacing')
        self.options.declare('span_cos_spacing', default=0,
                             desc='spanwise cosine spacing')
        self.options.declare('num_box_cp', default=0,
                             desc='number of chordwise CPs on the structural box')
        self.options.declare('num_twist_cp', default=0, desc='number of twist CPs')
        self.options.declare('S_ref_type', default='wetted',
                             desc='type of computed wing area (wetted or projected)')
        self.options.declare('fem_model_type', default='wingbox',
                             desc='type of FEM model (wingbox or tube)')
        self.options.declare('with_viscous', default=True, desc='viscous drag selection')
        self.options.declare('with_wave', default=True, desc='wave drag selection')
        self.options.declare('k_lam', default=0.05,
                             desc='fraction of chord with laminar flow')
        self.options.declare('c_max_t', default=0.38,
                             desc='chordwise location of maximum thickness')
        self.options.declare('E', default=73.1e9, desc='Youngs Modulus for AL 7073 [Pa]')
        self.options.declare('G', default=(73.1e9 / 2 / 1.33),
                             desc='Shear Modulus for AL 7073 [Pa]')
        self.options.declare('yield', default=(420.0e6 / 1.5),
                             desc='Allowable yield stress for AL 7073 [Pa]')
        self.options.declare('mrho', default=2.78e3,
                             desc='Material density for AL 7073 [kg/m^3]')
        self.options.declare('strength_factor_for_upper_skin', default=1.0,
                             desc='the yield stress is multiplied by this factor for the upper skin')
        self.options.declare('wing_weight_ratio', default=1.00,
                             desc='Ratio of the total wing weight (including non-structural components) to the wing structural weight.')
        self.options.declare('exact_failure_constraint',
                             default=False, desc='if false, use KS function')
        self.options.declare('struct_weight_relief', default=True,
                             desc='if true, use structural weight as intertia relief loads')
        self.options.declare('distributed_fuel_weight', default=True,
                             desc='Set True to distribute the fuel weight across the entire wing.')
        self.options.declare('n_point_masses', default=0, desc='number of point masses')
        self.options.declare('fuel_density', default=803.0,
                             desc='fuel density (only needed if the fuel-in-wing volume constraint is used) [kg/m^3]')

    def setup(self):
        self.add_input('box_upper_x', units='unitless',
                       shape=(self.options['num_box_cp'],))
        self.add_input('box_lower_x', units='unitless',
                       shape=(self.options['num_box_cp'],))
        self.add_input('box_upper_y', units='unitless',
                       shape=(self.options['num_box_cp'],))
        self.add_input('box_lower_y', units='unitless',
                       shape=(self.options['num_box_cp'],))
        self.add_input('twist_cp', units='deg', shape=(self.options['num_twist_cp'],))
        self.add_input('spar_thickness_cp', units='m',
                       shape=(self.options['num_twist_cp'],))
        self.add_input('skin_thickness_cp', units='m',
                       shape=(self.options['num_twist_cp'],))
        self.add_input('t_over_c_cp', units='unitless',
                       shape=(self.options['num_twist_cp'],))
        self.add_input('airfoil_t_over_c', units='unitless')
        self.add_input('fuel', val=0.0, units='kg')
        self.add_input('fuel_reserve', val=0.0, units='kg')
        self.add_input('CL0', val=0.0, units='unitless')
        self.add_input('CD0', val=0.0, units='unitless')
        self.add_input('cruise_Mach', val=0.0, units='unitless')
        self.add_input('cruise_altitude', val=0.0, units='m')
        self.add_input('cruise_range', val=0.0, units='m')
        self.add_input('cruise_SFC', val=0.0, units='1/s')
        self.add_input('engine_mass', val=0.0, units='kg')
        self.add_input('engine_location', val=np.array([0.0, 0.0, 0.0]), units='m')

        self.add_output('wing_weight', units='kg')
        self.add_output('fuel_burn', units='kg')

        self.declare_partials(of=['*'], wrt=['fuel'], method='fd')

        self.previous_DV_values = {}

    def compute(self, inputs, outputs):

        start_time = time.time()

        # perform the wing structural optimization and return the wing weight

        mesh = user_mesh()

        # surface options dictionary
        surf_dict = {
            # surface name
            'name': 'meshwing',

            # wing symmetry
            'symmetry': self.options['symmetry'],

            # wing mesh
            'mesh': mesh,

            # wing definition
            'S_ref_type': self.options['S_ref_type'],
            'fem_model_type': self.options['fem_model_type'],

            # wing thickness data
            'data_x_upper': inputs['box_upper_x'],
            'data_x_lower': inputs['box_lower_x'],
            'data_y_upper': inputs['box_upper_y'],
            'data_y_lower': inputs['box_lower_y'],

            # wing sizing parameters
            'twist_cp': inputs['twist_cp'],
            'spar_thickness_cp': inputs['spar_thickness_cp'],
            'skin_thickness_cp': inputs['skin_thickness_cp'],
            't_over_c_cp': inputs['t_over_c_cp'],
            'original_wingbox_airfoil_t_over_c': inputs['airfoil_t_over_c'],

            # Aerodynamic deltas.
            # These CL0 and CD0 values are added to the CL and CD
            # obtained from aerodynamic analysis of the surface to get
            # the total CL and CD.
            # These CL0 and CD0 values do not vary wrt alpha.
            # They can be used to account for things that are not included, such as contributions from the fuselage, camber, etc.
            'CL0': inputs['CL0'][0],
            'CD0': inputs['CD0'][0],
            'with_viscous': self.options['with_viscous'],
            'with_wave': self.options['with_wave'],

            # Airfoil properties for viscous drag calculation
            'k_lam': self.options['k_lam'],

            # flow, used for viscous drag
            'c_max_t': self.options['c_max_t'],

            # Structural values
            'E': self.options['E'],
            'G': self.options['G'],
            'yield': self.options['yield'],
            'mrho': self.options['mrho'],
            'strength_factor_for_upper_skin': self.options['strength_factor_for_upper_skin'],
            'wing_weight_ratio': self.options['wing_weight_ratio'],
            'exact_failure_constraint': self.options['exact_failure_constraint'],

            # structural weight factors
            'struct_weight_relief': self.options['struct_weight_relief'],
            'distributed_fuel_weight': self.options['distributed_fuel_weight'],

            # point masses
            'n_point_masses': self.options['n_point_masses'],

            # fuel factors
            'fuel_density': self.options['fuel_density'],
            'Wf_reserve': inputs['fuel_reserve'][0],
        }

        # define the surfaces
        surfaces = [surf_dict]

        # Create the problem and assign the model group
        prob = om.Problem()

        fuel = inputs['fuel'][0]

        mach = np.array([inputs['cruise_Mach'][0], 0.64])
        altitude = np.array([inputs['cruise_altitude'][0], 0.0])
        cruise_range = inputs['cruise_range'][0]

        atmos = Atmosphere(altitude)

        speed_of_sound = atmos.speed_of_sound
        rho = atmos.density
        dynamic_viscosity = atmos.dynamic_viscosity

        velocity = np.array([mach[0] * speed_of_sound[0], mach[1] * speed_of_sound[1]])

        # set values on the subproblem based on what's passed in from Aviary
        # Add problem information as an independent variables component
        indep_var_comp = om.IndepVarComp()
        indep_var_comp.add_output('Mach_number', val=mach,
                                  desc='Mach number for cruise and for maneuver')
        indep_var_comp.add_output('v', val=velocity, units='m/s',
                                  desc='velocity for cruise and for maneuver')
        indep_var_comp.add_output('re',
                                  val=np.array([rho[0] * velocity[0] * 1.0 / (dynamic_viscosity[0]),
                                                rho[1] * velocity[1] * 1.0 / (dynamic_viscosity[1])]),
                                  desc='Reynolds number per unit length for cruise and for maneuver', units='1/m'
                                  )
        indep_var_comp.add_output('rho', val=rho, units='kg/m**3',
                                  desc='atmostheric density for cruise and for maneuver')
        indep_var_comp.add_output('speed_of_sound', val=speed_of_sound,
                                  units='m/s', desc='speed of sound for cruise and for maneuver')
        indep_var_comp.add_output('CT', val=0.53 / 3600, units='1/s',
                                  desc='cruise thrust specific fuel consumption')
        indep_var_comp.add_output('R', val=cruise_range, units='m', desc='cruise range')
        indep_var_comp.add_output('W0_without_point_masses',
                                  val=40.e3 + fuel + surf_dict['Wf_reserve'], units='kg')
        indep_var_comp.add_output('load_factor', val=np.array(
            [1.0, 2.5]), desc='load factor for cruise and for maneuver')
        indep_var_comp.add_output('alpha', val=0.0, units='deg')
        indep_var_comp.add_output('alpha_maneuver', val=0.0, units='deg')
        indep_var_comp.add_output('empty_cg', val=np.zeros((3)), units='m')
        indep_var_comp.add_output('fuel_mass', val=fuel, units='kg')

        # add the problem variables subsystem
        prob.model.add_subsystem('prob_vars', indep_var_comp, promotes=['*'])

        # point masses
        point_masses = inputs['engine_mass']
        point_mass_locations = inputs['engine_location']

        indep_var_comp.add_output('point_masses', val=point_masses, units='kg')
        indep_var_comp.add_output('point_mass_locations',
                                  val=point_mass_locations, units='m')

        # add an ExecComp subsystem to compute the actual W0 to be used within OAS based on the sum of the point mass and other W0 weight
        prob.model.add_subsystem(
            'W0_comp', om.ExecComp('W0 = W0_without_point_masses + 2 * sum(point_masses)', units='kg'), promotes=['*']
        )

        # Loop over each surface in the surfaces list
        for surface in surfaces:
            # Get the surface name and create a group to contain components only for this surface
            name = surface['name']

            aerostruct_group = AerostructGeometry(surface=surface)

            # Add groups to the problem with the name of the surface.
            prob.model.add_subsystem(name, aerostruct_group)

        # Loop through and add the specified number of aerostruct points
        for i in range(2):
            point_name = 'AS_point_{}'.format(i)

            # Create the aerostruct point group and add it to the model
            AS_point = AerostructPoint(
                surfaces=surfaces, internally_connect_fuelburn=False)
            prob.model.add_subsystem(point_name, AS_point)

            # Connect flow properties to the analysis point
            prob.model.connect('v', point_name + '.v', src_indices=[i])
            prob.model.connect('Mach_number', point_name +
                               '.Mach_number', src_indices=[i])
            prob.model.connect('re', point_name + '.re', src_indices=[i])
            prob.model.connect('rho', point_name + '.rho', src_indices=[i])
            prob.model.connect('CT', point_name + '.CT')
            prob.model.connect('R', point_name + '.R')
            prob.model.connect('W0', point_name + '.W0')
            prob.model.connect('speed_of_sound', point_name +
                               '.speed_of_sound', src_indices=[i])
            prob.model.connect('empty_cg', point_name + '.empty_cg')
            prob.model.connect('load_factor', point_name +
                               '.load_factor', src_indices=[i])
            prob.model.connect('fuel_mass', point_name +
                               '.total_perf.L_equals_W.fuelburn')
            prob.model.connect('fuel_mass', point_name + '.total_perf.CG.fuelburn')

            for surface in surfaces:
                name = surface['name']

                if surf_dict['distributed_fuel_weight']:
                    prob.model.connect('load_factor', point_name +
                                       '.coupled.load_factor', src_indices=[i])

                com_name = point_name + '.' + name + '_perf.'
                prob.model.connect(
                    name + '.local_stiff_transformed', point_name +
                    '.coupled.' + name + '.local_stiff_transformed'
                )

                prob.model.connect(name + '.nodes', point_name +
                                   '.coupled.' + name + '.nodes')

                # Connect aerodyamic mesh to coupled group mesh
                prob.model.connect(name + '.mesh', point_name +
                                   '.coupled.' + name + '.mesh')

                if surf_dict['struct_weight_relief']:
                    prob.model.connect(name + '.element_mass',
                                       point_name + '.coupled.' + name + '.element_mass')

                # Connect performance calculation variables
                prob.model.connect(name + '.nodes', com_name + 'nodes')
                prob.model.connect(name + '.cg_location', point_name +
                                   '.' + 'total_perf.' + name + '_cg_location')
                prob.model.connect(name + '.structural_mass', point_name +
                                   '.' + 'total_perf.' + name + '_structural_mass')

                # Connect wingbox properties to von Mises stress calcs
                prob.model.connect(name + '.Qz', com_name + 'Qz')
                prob.model.connect(name + '.J', com_name + 'J')
                prob.model.connect(name + '.A_enc', com_name + 'A_enc')
                prob.model.connect(name + '.htop', com_name + 'htop')
                prob.model.connect(name + '.hbottom', com_name + 'hbottom')
                prob.model.connect(name + '.hfront', com_name + 'hfront')
                prob.model.connect(name + '.hrear', com_name + 'hrear')

                prob.model.connect(name + '.spar_thickness', com_name + 'spar_thickness')
                prob.model.connect(name + '.t_over_c', com_name + 't_over_c')

                coupled_name = point_name + '.coupled.' + name
                prob.model.connect('point_masses', coupled_name + '.point_masses')
                prob.model.connect('point_mass_locations',
                                   coupled_name + '.point_mass_locations')

        # use only the first surface for constraints
        surface = surfaces[0]
        name = surface['name']

        prob.model.connect('alpha', 'AS_point_0' + '.alpha')
        prob.model.connect('alpha_maneuver', 'AS_point_1' + '.alpha')

        # Here we add the fuel volume constraint componenet to the model
        prob.model.add_subsystem(
            'fuel_vol_delta', WingboxFuelVolDelta(surface=surfaces[0]))
        prob.model.connect(name + '.struct_setup.fuel_vols', 'fuel_vol_delta.fuel_vols')
        prob.model.connect('AS_point_0.fuelburn', 'fuel_vol_delta.fuelburn')

        if surf_dict['distributed_fuel_weight']:
            prob.model.connect(name + '.struct_setup.fuel_vols',
                               'AS_point_0.coupled.' + name + '.struct_states.fuel_vols')
            prob.model.connect('fuel_mass', 'AS_point_0.coupled.' +
                               name + '.struct_states.fuel_mass')

            prob.model.connect(name + '.struct_setup.fuel_vols',
                               'AS_point_1.coupled.' + name + '.struct_states.fuel_vols')
            prob.model.connect('fuel_mass', 'AS_point_1.coupled.' +
                               name + '.struct_states.fuel_mass')

        # add an ExecComp to compute the fuel difference
        comp = om.ExecComp('fuel_diff = (fuel_mass - fuelburn) / fuelburn', units='kg')

        # add a fuel difference subsystem
        prob.model.add_subsystem('fuel_diff', comp, promotes_inputs=[
                                 'fuel_mass'], promotes_outputs=['fuel_diff'])
        prob.model.connect('AS_point_0.fuelburn', 'fuel_diff.fuelburn')

        # add an objective funtion
        prob.model.add_objective('AS_point_0.fuelburn', scaler=1e-5)

        # add design variables
        prob.model.add_design_var(name + '.twist_cp', lower=-
                                  15.0, upper=15.0, scaler=0.1)
        prob.model.add_design_var(name + '.spar_thickness_cp',
                                  lower=0.003, upper=0.1, scaler=1e2)
        prob.model.add_design_var(name + '.skin_thickness_cp',
                                  lower=0.003, upper=0.1, scaler=1e2)
        prob.model.add_design_var(name + '.geometry.t_over_c_cp',
                                  lower=0.07, upper=0.2, scaler=10.0)
        prob.model.add_design_var('alpha_maneuver', lower=-15.0, upper=15)

        # add problem constraints
        prob.model.add_constraint('AS_point_0.CL', equals=0.5)
        prob.model.add_constraint('AS_point_1.L_equals_W', equals=0.0)
        prob.model.add_constraint('AS_point_1.' + name + '_perf.failure', upper=0.0)

        prob.model.add_constraint('fuel_vol_delta.fuel_vol_delta', lower=0.0)

        prob.model.add_design_var('fuel_mass', lower=0.0, upper=2e5, scaler=1e-5)
        prob.model.add_constraint('fuel_diff', equals=0.0)

        # set up the optimization
        prob.driver = om.ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        prob.driver.options['tol'] = 1e-8

        # Set up the problem
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=om.PromotionWarning)
            prob.setup()

        # change linear solver for aerostructural coupled adjoint
        prob.model.AS_point_0.coupled.linear_solver = om.LinearBlockGS(
            iprint=0, maxiter=30, use_aitken=True)
        prob.model.AS_point_1.coupled.linear_solver = om.LinearBlockGS(
            iprint=0, maxiter=30, use_aitken=True)
        prob.model.AS_point_0.coupled.nonlinear_solver.options['iprint'] = 0
        prob.model.AS_point_1.coupled.nonlinear_solver.options['iprint'] = 0

        # Loop through self.previous_DV_values and set each prob value
        for key, value in self.previous_DV_values.items():
            prob[key] = value

        # run the problem
        prob.run_driver()

        self.previous_DV_values[name + '.twist_cp'] = prob[name + '.twist_cp']
        self.previous_DV_values[name +
                                '.spar_thickness_cp'] = prob[name + '.spar_thickness_cp']
        self.previous_DV_values[name +
                                '.skin_thickness_cp'] = prob[name + '.skin_thickness_cp']
        self.previous_DV_values[name +
                                '.geometry.t_over_c_cp'] = prob[name + '.geometry.t_over_c_cp']
        self.previous_DV_values['alpha_maneuver'] = prob['alpha_maneuver']

        # output wing weight and fuel burn
        outputs['wing_weight'] = prob[name + '.structural_mass'][0]
        outputs['fuel_burn'] = prob['AS_point_0.fuelburn'][0]

        # calculate execution time
        delta_time = time.time() - start_time
        tm_hrs, remainder = divmod(delta_time, 3600)
        tm_min, remainder = divmod(remainder, 60)
        tm_sec = int(remainder)
        tm_msec = (remainder - tm_sec) * 1000
        print('Structures OAS Compute End --- execution time {:02}:{:02}:{:02}.{:03}'
              .format(int(tm_hrs), int(tm_min), int(tm_sec), int(tm_msec)))


if __name__ == "__main__":
    # run program
    print("OAS optimization")
    prob = om.Problem()

    # mesh example
    prob.model.add_subsystem('OAS', OAStructures(
        symmetry=True,
        wing_weight_ratio=1.0,
        S_ref_type='projected',
        n_point_masses=1,
        num_twist_cp=4,
        num_box_cp=51))

    prob.setup()

    # test data taken from the OpenAeroStruct example aeroelastic wingbox example
    # and the aircraft_for_bench_FwFm.csv benchmark data file.  All length units are in meters
    # and all mass units are in kilograms for this test data.
    prob['OAS.box_upper_x'] = np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.26, 0.27, 0.28, 0.29, 0.3, 0.31, 0.32,
                                       0.33, 0.34, 0.35, 0.36, 0.37, 0.38, 0.39, 0.4, 0.41, 0.42, 0.43, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.5, 0.51, 0.52, 0.53, 0.54, 0.55, 0.56, 0.57, 0.58, 0.59, 0.6])
    prob['OAS.box_lower_x'] = np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15, 0.16, 0.17, 0.18, 0.19, 0.2, 0.21, 0.22, 0.23, 0.24, 0.25, 0.26, 0.27, 0.28, 0.29, 0.3, 0.31, 0.32,
                                       0.33, 0.34, 0.35, 0.36, 0.37, 0.38, 0.39, 0.4, 0.41, 0.42, 0.43, 0.44, 0.45, 0.46, 0.47, 0.48, 0.49, 0.5, 0.51, 0.52, 0.53, 0.54, 0.55, 0.56, 0.57, 0.58, 0.59, 0.6])
    prob['OAS.box_upper_y'] = np.array([0.0447,  0.046,  0.0472,  0.0484,  0.0495,  0.0505,  0.0514,  0.0523,  0.0531,  0.0538, 0.0545,  0.0551,  0.0557, 0.0563,  0.0568, 0.0573,  0.0577,  0.0581,  0.0585,  0.0588,  0.0591,  0.0593,  0.0595,  0.0597,
                                       0.0599,  0.06,    0.0601,  0.0602,  0.0602,  0.0602,  0.0602,  0.0602,  0.0601,  0.06,    0.0599,  0.0598,  0.0596,  0.0594,  0.0592,  0.0589,  0.0586,  0.0583,  0.058,   0.0576,  0.0572,  0.0568,  0.0563,  0.0558,  0.0553,  0.0547,  0.0541])
    prob['OAS.box_lower_y'] = np.array([-0.0447, -0.046, -0.0473, -0.0485, -0.0496, -0.0506, -0.0515, -0.0524, -0.0532, -0.054, -0.0547, -0.0554, -0.056, -0.0565, -0.057, -0.0575, -0.0579, -0.0583, -0.0586, -0.0589, -0.0592, -0.0594, -0.0595, -0.0596, -
                                       0.0597, -0.0598, -0.0598, -0.0598, -0.0598, -0.0597, -0.0596, -0.0594, -0.0592, -0.0589, -0.0586, -0.0582, -0.0578, -0.0573, -0.0567, -0.0561, -0.0554, -0.0546, -0.0538, -0.0529, -0.0519, -0.0509, -0.0497, -0.0485, -0.0472, -0.0458, -0.0444])
    prob['OAS.twist_cp'] = np.array([-6., -6., -4., 0.])
    prob['OAS.spar_thickness_cp'] = np.array([0.004, 0.005, 0.008, 0.01])
    prob['OAS.skin_thickness_cp'] = np.array([0.005, 0.01, 0.015, 0.025])
    prob['OAS.t_over_c_cp'] = np.array([0.08, 0.08, 0.10, 0.08])
    prob['OAS.airfoil_t_over_c'] = 0.13
    prob['OAS.fuel'] = 18163.652864
    prob['OAS.fuel_reserve'] = 1360.77711
    prob['OAS.CD0'] = 0.0078
    prob['OAS.cruise_Mach'] = 0.785
    prob['OAS.cruise_altitude'] = 11303.682962301647
    prob['OAS.cruise_range'] = 6482000.0
    prob['OAS.cruise_SFC'] = 0.53 / 3600
    prob['OAS.engine_mass'] = 3356.583538
    prob['OAS.engine_location'] = np.array([4.825, -1.0, 0.0])

    prob.run_model()

    print('wing weight = ', prob.model.get_val('OAS.wing_weight', units='lbm'))
    print('fuel burn = ', prob.model.get_val('OAS.fuel_burn', units='lbm'))



import openmdao.api as om

from aviary.subsystems.mass.flops_based.air_conditioning import (
    AltAirCondMass, TransportAirCondMass)
from aviary.subsystems.mass.flops_based.anti_icing import AntiIcingMass
from aviary.subsystems.mass.flops_based.apu import TransportAPUMass
from aviary.subsystems.mass.flops_based.avionics import TransportAvionicsMass
from aviary.subsystems.mass.flops_based.canard import CanardMass
from aviary.subsystems.mass.flops_based.cargo import CargoMass
from aviary.subsystems.mass.flops_based.cargo_containers import TransportCargoContainersMass
from aviary.subsystems.mass.flops_based.crew import NonFlightCrewMass, FlightCrewMass
from aviary.subsystems.mass.flops_based.electrical import (
    AltElectricalMass, ElectricalMass)
from aviary.subsystems.mass.flops_based.engine import EngineMass
from aviary.subsystems.mass.flops_based.engine_controls import TransportEngineCtrlsMass
from aviary.subsystems.mass.flops_based.engine_oil import (
    TransportEngineOilMass, AltEngineOilMass)
from aviary.subsystems.mass.flops_based.fin import FinMass
from aviary.subsystems.mass.flops_based.fuel_capacity import FuelCapacityGroup
from aviary.subsystems.mass.flops_based.fuel_system import (
    AltFuelSystemMass, TransportFuelSystemMass)
from aviary.subsystems.mass.flops_based.furnishings import (
    AltFurnishingsGroupMass, AltFurnishingsGroupMassBase,
    TransportFurnishingsGroupMass)
from aviary.subsystems.mass.flops_based.fuselage import (
    AltFuselageMass, TransportFuselageMass)
from aviary.subsystems.mass.flops_based.horizontal_tail import (
    AltHorizontalTailMass, HorizontalTailMass)
from aviary.subsystems.mass.flops_based.hydraulics import (
    AltHydraulicsGroupMass, TransportHydraulicsGroupMass)
from aviary.subsystems.mass.flops_based.instruments import TransportInstrumentMass
from aviary.subsystems.mass.flops_based.landing_group import LandingMassGroup
from aviary.subsystems.mass.flops_based.mass_summation import MassSummation
from aviary.subsystems.mass.flops_based.misc_engine import EngineMiscMass
from aviary.subsystems.mass.flops_based.nacelle import NacelleMass
from aviary.subsystems.mass.flops_based.paint import PaintMass
from aviary.subsystems.mass.flops_based.passenger_service import (
    AltPassengerServiceMass, PassengerServiceMass)
from aviary.subsystems.mass.flops_based.starter import TransportStarterMass
from aviary.subsystems.mass.flops_based.surface_controls import (
    AltSurfaceControlMass, SurfaceControlMass)
from aviary.subsystems.mass.flops_based.thrust_reverser import ThrustReverserMass
from aviary.subsystems.mass.flops_based.unusable_fuel import (
    AltUnusableFuelMass, TransportUnusableFuelMass)
from aviary.subsystems.mass.flops_based.vertical_tail import (
    AltVerticalTailMass, VerticalTailMass)
from aviary.subsystems.mass.flops_based.wing_group import WingMassGroup
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Mission


class MassPremission(om.Group):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        alt_mass = aviary_options.get_val(Aircraft.Design.USE_ALT_MASS)

        self.add_subsystem(
            'cargo',
            CargoMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'cargo_containers',
            TransportCargoContainersMass(
                aviary_options=aviary_options),
            promotes_inputs=['*', ], promotes_outputs=['*', ])

        self.add_subsystem(
            'engine_controls',
            TransportEngineCtrlsMass(aviary_options=aviary_options),
            promotes_inputs=['*', ], promotes_outputs=['*', ])

        self.add_subsystem(
            'avionics',
            TransportAvionicsMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'fuel_capacity_group',
            FuelCapacityGroup(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'engine_mass',
            EngineMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        if alt_mass:
            self.add_subsystem(
                'fuel_system',
                AltFuelSystemMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'AC',
                AltAirCondMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'engine_oil',
                AltEngineOilMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'furnishing_base',
                AltFurnishingsGroupMassBase(
                    aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'furnishings',
                AltFurnishingsGroupMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'hydraulics',
                AltHydraulicsGroupMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'pass_service',
                AltPassengerServiceMass(
                    aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'unusable_fuel',
                AltUnusableFuelMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'electrical',
                AltElectricalMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        else:
            self.add_subsystem(
                'fuel_system',
                TransportFuelSystemMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'AC',
                TransportAirCondMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'engine_oil',
                TransportEngineOilMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'furnishings',
                TransportFurnishingsGroupMass(
                    aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'hydraulics',
                TransportHydraulicsGroupMass(
                    aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'pass_service',
                PassengerServiceMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'unusable_fuel',
                TransportUnusableFuelMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'electrical',
                ElectricalMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'starter',
            TransportStarterMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'anti_icing',
            AntiIcingMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'apu',
            TransportAPUMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'nonflight_crew',
            NonFlightCrewMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'flight_crew',
            FlightCrewMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'instruments',
            TransportInstrumentMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'misc_engine',
            EngineMiscMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'nacelle',
            NacelleMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'paint',
            PaintMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'thrust_rev',
            ThrustReverserMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'landing_group',
            LandingMassGroup(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        if alt_mass:
            self.add_subsystem(
                'surf_ctrl',
                AltSurfaceControlMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'fuselage',
                AltFuselageMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'htail',
                AltHorizontalTailMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'vert_tail',
                AltVerticalTailMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        else:
            self.add_subsystem(
                'surf_ctrl',
                SurfaceControlMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

            self.add_subsystem(
                'fuselage',
                TransportFuselageMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'htail',
                HorizontalTailMass(aviary_options=aviary_options),
                promotes_inputs=['*', ], promotes_outputs=['*', ])

            self.add_subsystem(
                'vert_tail',
                VerticalTailMass(aviary_options=aviary_options),
                promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'canard',
            CanardMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'fin',
            FinMass(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'wing_group',
            WingMassGroup(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])

        self.add_subsystem(
            'total_mass',
            MassSummation(aviary_options=aviary_options),
            promotes_inputs=['*'], promotes_outputs=['*'])



from abc import ABC

from aviary.variable_info.variable_meta_data import _MetaData


class SubsystemBuilderBase(ABC):
    __slots__ = ('name', 'meta_data')

    # derived type customization point
    default_name = 'default_subsystem_name'

    def __init__(self, name=None, meta_data=None):
        if name is None:
            name = self.default_name
        self.name = name

        if meta_data is None:
            meta_data = _MetaData
        self.meta_data = meta_data

    def build_pre_mission(self, aviary_inputs, **kwargs):
        """
        Build an OpenMDAO System for the pre-mission computations of the subsystem.

        Required for subsystems with pre-mission computations.

        Used in level3.py to build the pre-mission system.

        Returns
        -------
        pre_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen in
            the pre-mission (formerly statics) part of the Aviary problem. This
            includes sizing, design, and other non-mission parameters.
        """
        return None

    def get_states(self):
        """
        Return a dictionary of states defined by this subsystem.

        Required for subsystems with mission-based dynamics.

        Note that there must be outputs provided by the user's model that provide
        the time derivative of each state. The recommended convention is to name the output
        f"{state_name}_rate", where `state_name` is the name of the state variable.

        Use in the phase builders (e.g. cruise_phase.py) when other states are added to the phase.

        Returns
        -------
        states : dict
            A dictionary where the keys are the names of the state variables
            and the values are dictionaries with the following keys:

            - 'rate_source': a string indicating the name of the output that provides the time derivative of the state variable
            - 'units': a string indicating the units of the state variable
            - any additional keyword arguments required by Dymos for the state variable.
        """
        return {}

    def get_controls(self, phase_name=None):
        """
        Return a dictionary of control variables for the subsystem.

        Parameters
        ----------
        phase_name : str, optional
            Name of the flight phase. This allows for different control variables to be used in different flight phases.
            You can add branching logic however you want based on the phase_name within a builder.

        Notes
        -----
        This method is optional, used if subsystems have control variables.
        Used in the phase builders (e.g. cruise_phase.py) when other controls are added to the phase.

        Returns
        -------
        controls : dict
            A dictionary where the keys are the names of the control variables
            and the values are dictionaries with the following keys:

            - 'units': str
                The units for the control variable.
            - any additional keyword arguments required by OpenMDAO for the control
            variable.
        """
        return {}

    def get_parameters(self):
        """
        Return a dictionary of fixed values for the subsystem.

        Optional, used if subsystems have fixed values.

        Used in the phase builders (e.g. cruise_phase.py) when other parameters are added to the phase.

        This is distinct from `get_design_vars` in a nuanced way. Design variables
        are variables that are optimized by the problem that are not at the phase level.
        An example would be something that occurs in the pre-mission level of the problem.
        Parameters are fixed values that are held constant throughout a phase, but if
        `opt=True`, they are able to change during the optimization.

        Returns
        -------
        fixed_values : dict
            A dictionary where the keys are the names of the fixed variables
            and the values are dictionaries with the following keys:

            - 'value': float or array
                The fixed value for the variable.
            - 'units': str
                The units for the fixed value (optional).
            - any additional keyword arguments required by OpenMDAO for the fixed
              variable.
        """
        return {}

    def get_constraints(self):
        """
        Return a dictionary of constraints for the subsystem.

        Optional, used if subsystems have path or boundary constraints.

        Used in the phase builders (e.g. cruise_phase.py) when other constraints are added to the phase.

        Returns
        -------
        constraints : dict
            A dictionary where the keys are the names of the constraint variables
            and the values are dictionaries with the following keys:

            - type : str
                The type of constraint. Must be one of 'path' or 'boundary'.
            - any additional keyword arguments required by OpenMDAO for the constraint
              variable.
        """
        return {}

    def get_linked_variables(self):
        """
        Return a list of variable names that will be linked between phases.

        Optional

        Returns
        -------
        linked_vars : list of variables to link between phases
        """
        return []

    def get_bus_variables(self):
        """
        Return a dictionary of variables that will be passed from the pre-mission
        to mission systems.

        Returns
        -------
        bus_variables : dict
            A dictionary where the keys are the names of the variables
            output in the pre-mission subsystem and the values are 
            dictionaries with the following keys:

            - 'mission_name': str or list
                Names of the input variable to be connected in the mission
                subsystem (optional).
            - 'post_mission_name': str or list
                Names of the input variable to be connected in the post-
                mission subsystem (optional).
            - 'units' : str
                This is temporary and will be removed, however, it is
                currently a requirement. 

        """
        return {}

    def build_mission(self, num_nodes, aviary_inputs, **kwargs):
        """
        Build an OpenMDAO System for the mission computations of the subsystem.

        Required for subsystems with mission-based dynamics.

        Used in the ODE class definition (e.g. mission_ODE.py) to build the mission system.

        Returns
        -------
        mission_sys : openmdao.core.System
            An OpenMDAO System containing all computations that need to happen
            during the mission. This includes time-dependent states that are
            being integrated as well as any other variables that vary during
            the mission.
        """
        return None

    def define_order(self):
        """
        Return a list of subsystem names that must be defined before this one. E.g., must go before or after aero or prop.

        Optional

        This will be made easier once an upcoming POEM about reordering systems in OM is integrated.

        Not needed yet, but will be useful when subsystem order matters, like for certain aero-prop cases.

        Returns
        -------
        order : list of str
            A list of subsystem names that this subsystem depends on.
        """
        return []

    def get_design_vars(self):
        """
        Return a dictionary of design variables for the subsystem.

        Optional

        Not currently used.

        Returns
        -------
        design_vars : dict
            A dictionary where the keys are the names of the design variables
            and the values are dictionaries with the following keys:

            - any additional keyword arguments required by OpenMDAO for the design
              variable.
        """
        return {}

    def get_initial_guesses(self):
        """
        Return a dictionary of initial guesses for the subsystem.

        Optional

        Returns
        -------
        initial_guesses : dict
            A dictionary where the keys are the names of the variables
            and the values are dictionaries with the following keys:

            - 'val': float or array
                The initial guess for the variable.
            - 'units': str
                The units for the initial guess (optional).
            - any additional keyword arguments required by OpenMDAO for the initial
              guess.
        """
        return {}

    def get_mass_names(self):
        """
        Return a list of names of the mass variables for the subsystem.

        Returns
        -------
        mass_names : list of str
            A list of the names of the mass variables for the subsystem.
        """
        return []

    def preprocess_inputs(self, aviary_inputs):
        """
        Preprocess the inputs to the subsystem, returning a modified AviaryValues object.

        This method is called after the inputs are passed to the subsystem. It
        can be used to modify the inputs before they are used in the subsystem.

        Parameters
        ----------
        aviary_inputs : dict
            A dictionary containing the inputs to the subsystem.
        """
        return aviary_inputs

    def get_outputs(self):
        """
        Return a list of output variable names for the subsystem. Also adds
        all of these outputs to the Dymos timeseries outputs for use when graphing
        or post-processing the mission.

        Useful for knowing what a subsystem provides that other subsystems can use.
        """
        return []

    def build_post_mission(self, aviary_inputs, **kwargs):
        """
        Build an OpenMDAO System for the post-mission computations of the subsystem.

        Required for subsystems with post-mission-based analyses.

        Returns
        -------
        post_mission_sys : openmdao.core.System
            An OpenMDAO system containing all computations that need to happen
            after the mission. This includes time-dependent states that are
            being integrated as well as any other variables that vary after
            the mission.
        """
        return None

    def report(self, prob, reports_folder, **kwargs):
        """
        Generates report file for this subsystem. If this subsystem doesn't need a
        report, do nothing.

        Parameters
        ----------
        prob : AviaryProblem
            The AviaryProblem that will be used to generate the report
        reports_folder : Path
            Location of the subsystems_report folder this report will be placed in
        """
        return



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials

from aviary.subsystems.propulsion.propulsion_builder import CorePropulsionBuilder
from aviary.subsystems.geometry.geometry_builder import CoreGeometryBuilder
from aviary.subsystems.mass.mass_builder import CoreMassBuilder
from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.subsystems.premission import CorePreMission
from aviary.utils.aviary_values import get_items, get_keys
from aviary.validation_cases.validation_tests import (
    get_flops_case_names, get_flops_inputs, get_flops_outputs
)
from aviary.variable_info.variables import Aircraft, Mission
from aviary.variable_info.variable_meta_data import _MetaData as BaseMetaData
from aviary.models.large_single_aisle_1.V3_bug_fixed_IO import V3_bug_fixed_options, V3_bug_fixed_non_metadata
from aviary.utils.functions import set_aviary_initial_values
from aviary.variable_info.variables_in import VariablesIn
from aviary.variable_info.enums import LegacyCode


FLOPS = LegacyCode.FLOPS
GASP = LegacyCode.GASP

data_sets = get_flops_case_names()


def setup_options(priority_data, backfill_data):
    priority_keys = get_keys(priority_data)
    full_data = priority_data.deepcopy()

    # add options to priority data that exist in backfill data but not priorirty
    for (key, (val, units)) in get_items(backfill_data):
        if key not in priority_keys:
            (new_val, new_unit) = backfill_data.get_item(key)
            full_data.set_val(key, val=new_val, units=new_unit)

    return full_data


class PreMissionTestCase(unittest.TestCase):

    def setUp(self):

        # set up inputs such that GASP inputs take priority

        case_name = 'LargeSingleAisle1FLOPS'

        flops_inputs = get_flops_inputs(case_name)
        flops_outputs = get_flops_outputs(case_name)

        FLOPS_input = flops_inputs
        GASP_input = V3_bug_fixed_options

        self.prob = om.Problem()

        input_options = setup_options(GASP_input, FLOPS_input)

        # delete the options that would override values
        input_options.delete(Aircraft.Wing.AREA)
        input_options.delete(Aircraft.Fuselage.LENGTH)
        input_options.delete(Aircraft.Wing.SPAN)
        input_options.delete(Aircraft.Nacelle.AVG_DIAMETER)
        input_options.delete(Aircraft.Fuselage.WETTED_AREA)
        input_options.delete(Aircraft.Wing.ULTIMATE_LOAD_FACTOR)
        input_options.delete(Aircraft.Fuel.TOTAL_CAPACITY)
        input_options.delete(Aircraft.Nacelle.AVG_LENGTH)

        prop = CorePropulsionBuilder('core_propulsion', BaseMetaData)
        mass = CoreMassBuilder('core_mass', BaseMetaData, GASP)
        aero = CoreAerodynamicsBuilder('core_aerodynamics', BaseMetaData, FLOPS)
        geom = CoreGeometryBuilder('core_geometry',
                                   BaseMetaData,
                                   use_both_geometries=True,
                                   code_origin_to_prioritize=GASP)

        core_subsystems = [prop, geom, mass, aero]

        self.prob.model.add_subsystem(
            'pre_mission',
            CorePreMission(aviary_options=input_options,
                           subsystems=core_subsystems),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        # set defaults for all the vars
        val, units = input_options.get_item(Aircraft.Engine.SCALED_SLS_THRUST)
        self.prob.model.pre_mission.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=val, units=units)

        for (key, (val, units)) in get_items(GASP_input):
            try:
                if not BaseMetaData[key]['option']:
                    self.prob.model.set_input_defaults(key, val, units)
            except KeyError:
                continue

        for (key, (val, units)) in get_items(V3_bug_fixed_non_metadata):
            self.prob.model.set_input_defaults(key, val=val, units=units)

        self.prob.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=input_options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        set_aviary_initial_values(self.prob.model, input_options)
        self.prob.setup(check=False, force_alloc_complex=True)
        self.prob.set_solver_print(2)

        # Initial guess for gross mass.
        # We set it to an unconverged value to test convergence.
        self.prob.set_val(Mission.Design.GROSS_MASS, val=1000.0)

        # Set inital values for all variables.
        for (key, (val, units)) in get_items(input_options):
            try:
                self.prob.set_val(key, val, units)
            except KeyError:
                continue

    def test_GASP_mass_FLOPS_everything_else(self):
        self.prob.run_model()

        # check the outputs from GASP mass and geometry (FLOPS outputs are not tested)

        tol = 5e-4
        # size values:
        assert_near_equal(self.prob["gasp_based_geom.fuselage.cabin_height"], 13.1, tol)
        assert_near_equal(
            self.prob["gasp_based_geom.fuselage.cabin_len"], 72.09722222222223, tol)
        assert_near_equal(self.prob["gasp_based_geom.fuselage.nose_height"], 8.6, tol)

        assert_near_equal(self.prob[Aircraft.Wing.CENTER_CHORD], 17.63, tol)
        assert_near_equal(self.prob[Aircraft.Wing.ROOT_CHORD], 16.54, tol)
        assert_near_equal(
            self.prob[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED], 0.1397, tol
        )  # not exact GASP value from the output file, likely due to rounding error

        assert_near_equal(
            self.prob[Aircraft.HorizontalTail.AVERAGE_CHORD], 9.6509873673743, tol
        )
        assert_near_equal(
            self.prob[Aircraft.VerticalTail.AVERAGE_CHORD], 16.96457870166355, tol
        )
        assert_near_equal(self.prob[Aircraft.Nacelle.AVG_LENGTH], 14.7, tol)

        # fixed mass values:
        assert_near_equal(
            self.prob["fixed_mass.main_gear_mass"], 6384.35, tol)
        assert_near_equal(
            self.prob["fixed_mass.tail.loc_MAC_vtail"], 0.44959578484694906, tol
        )

        # wing values:
        assert_near_equal(
            self.prob["wing_mass.isolated_wing_mass"], 16205, tol)
        assert_near_equal(self.prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS], 12606, tol)
        assert_near_equal(self.prob[Aircraft.Engine.ADDITIONAL_MASS], 1765/2, tol)

        # fuel values:
        # modified from GASP value to account for updated crew mass. GASP value is
        # 78843.6
        assert_near_equal(
            self.prob["fuel_mass.fuel_and_oem.OEM_wingfuel_mass"], 78145.0, tol
        )
        # modified from GASP value to account for updated crew mass. GASP value is
        # 102408.05695930264
        assert_near_equal(
            self.prob["fuel_mass.fus_mass_full"], 102663.6529626, tol
        )
        # modified from GASP value to account for updated crew mass. GASP value is
        # 1757
        assert_near_equal(
            self.prob[Aircraft.Fuel.FUEL_SYSTEM_MASS], 1727.9, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 1757
        assert_near_equal(self.prob[Aircraft.Design.STRUCTURE_MASS], 50931.4, tol)
        assert_near_equal(
            self.prob[Aircraft.Fuselage.MASS], 18833.76678366, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 18814

        # modified from GASP value to account for updated crew mass. GASP value is
        # 42843.6
        assert_near_equal(
            self.prob[Mission.Design.FUEL_MASS_REQUIRED], 42145.0, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 42843.6
        assert_near_equal(
            self.prob[Aircraft.Propulsion.MASS], 16098.7, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 16127
        assert_near_equal(
            self.prob[Mission.Design.FUEL_MASS], 42145.0, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 42844.0
        assert_near_equal(
            self.prob["fuel_mass.fuel_mass_min"], 32105.0, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 32803.6
        assert_near_equal(
            self.prob[Aircraft.Fuel.WING_VOLUME_DESIGN], 842.53, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 856.4910800459031
        assert_near_equal(
            self.prob["fuel_mass.fuel_and_oem.OEM_fuel_vol"], 1562.21, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 1576.1710061411081
        assert_near_equal(
            self.prob[Aircraft.Design.OPERATING_MASS], 97255.0, tol
        )  # modified from GASP value to account for updated crew mass. GASP value is 96556.0
        # extra_fuel_mass calculated differently in this version, so test for fuel_mass.fuel_and_oem.payload_mass_max_fuel not included
        assert_near_equal(
            self.prob["fuel_mass.fuel_and_oem.volume_wingfuel_mass"], 57066.3, tol
        )
        assert_near_equal(
            self.prob["fuel_mass.max_wingfuel_mass"], 57066.3, tol)
        assert_near_equal(
            self.prob[Aircraft.Fuel.AUXILIARY_FUEL_CAPACITY], 0, tol
        )  # always zero when no body tank
        assert_near_equal(
            self.prob["fuel_mass.body_tank.extra_fuel_volume"], 0, tol
        )  # always zero when no body tank
        assert_near_equal(
            self.prob["fuel_mass.body_tank.max_extra_fuel_mass"], 0, tol
        )  # always zero when no body tank

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-10, rtol=1e-12)

    def test_manual_override(self):

        # Problem in setup is GASP prioritized, so shared inputs for FLOPS will be manually overriden.

        outs = self.prob.model.pre_mission.list_outputs(
            includes='*gasp*fuselage:avg_diam*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.gasp_based_geom.fuselage.parameters.{Aircraft.Fuselage.AVG_DIAMETER}')
        self.assertTrue('MANUAL_OVERRIDE' not in outs[0][1]['prom_name'])

        outs = self.prob.model.pre_mission.list_outputs(
            includes='*flops*fuselage:avg_diam*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.flops_based_geom.fuselage_prelim.{Aircraft.Fuselage.AVG_DIAMETER}')
        self.assertTrue('MANUAL_OVERRIDE' in outs[0][1]['prom_name'])

        outs = self.prob.model.pre_mission.list_outputs(
            includes='*gasp*fuselage:wetted_area*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.gasp_based_geom.fuselage.size.{Aircraft.Fuselage.WETTED_AREA}')
        self.assertTrue('MANUAL_OVERRIDE' not in outs[0][1]['prom_name'])

        outs = self.prob.model.pre_mission.list_outputs(
            includes='*flops*fuselage:wetted_area*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.flops_based_geom.fuselage.{Aircraft.Fuselage.WETTED_AREA}')
        self.assertTrue('MANUAL_OVERRIDE' in outs[0][1]['prom_name'])

        # Setup FLOPS prioritized problem.

        case_name = 'LargeSingleAisle1FLOPS'

        flops_inputs = get_flops_inputs(case_name)
        flops_outputs = get_flops_outputs(case_name)

        FLOPS_input = flops_inputs
        GASP_input = V3_bug_fixed_options

        prob = om.Problem()

        aviary_inputs = setup_options(GASP_input, FLOPS_input)

        aviary_inputs.delete(Aircraft.Fuselage.WETTED_AREA)

        prob = om.Problem()
        model = prob.model

        prop = CorePropulsionBuilder('core_propulsion', BaseMetaData)
        mass = CoreMassBuilder('core_mass', BaseMetaData, GASP)
        aero = CoreAerodynamicsBuilder('core_aerodynamics', BaseMetaData, FLOPS)
        geom = CoreGeometryBuilder('core_geometry',
                                   BaseMetaData,
                                   use_both_geometries=True,
                                   code_origin_to_prioritize=FLOPS)

        core_subsystems = [prop, geom, mass, aero]

        model.add_subsystem(
            'pre_mission',
            CorePreMission(aviary_options=aviary_inputs,
                           subsystems=core_subsystems),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        val, units = aviary_inputs.get_item(Aircraft.Engine.SCALED_SLS_THRUST)
        prob.model.pre_mission.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=val, units=units)

        for (key, (val, units)) in get_items(GASP_input):
            try:
                if not BaseMetaData[key]['option']:
                    prob.model.set_input_defaults(key, val, units)
            except KeyError:
                continue

        for (key, (val, units)) in get_items(V3_bug_fixed_non_metadata):
            prob.model.set_input_defaults(key, val=val, units=units)

        prob.model.add_subsystem(
            'input_sink',
            VariablesIn(aviary_options=aviary_inputs),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )

        set_aviary_initial_values(prob.model, aviary_inputs)

        prob.setup()

        # Problem in setup is FLOPS prioritized, so shared inputs for FLOPS will be manually overriden.

        outs = prob.model.pre_mission.list_outputs(
            includes='*gasp*fuselage:avg_diam*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.gasp_based_geom.fuselage.parameters.{Aircraft.Fuselage.AVG_DIAMETER}')
        self.assertTrue('MANUAL_OVERRIDE' in outs[0][1]['prom_name'])

        outs = prob.model.pre_mission.list_outputs(
            includes='*flops*fuselage:avg_diam*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.flops_based_geom.fuselage_prelim.{Aircraft.Fuselage.AVG_DIAMETER}')
        self.assertTrue('MANUAL_OVERRIDE' not in outs[0][1]['prom_name'])

        outs = prob.model.pre_mission.list_outputs(
            includes='*gasp*fuselage:wetted_area*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.gasp_based_geom.fuselage.size.{Aircraft.Fuselage.WETTED_AREA}')
        self.assertTrue('MANUAL_OVERRIDE' in outs[0][1]['prom_name'])

        outs = prob.model.pre_mission.list_outputs(
            includes='*flops*fuselage:wetted_area*', prom_name=True)

        self.assertTrue(
            outs[0][0] == f'core_geometry.flops_based_geom.fuselage.{Aircraft.Fuselage.WETTED_AREA}')
        self.assertTrue('MANUAL_OVERRIDE' not in outs[0][1]['prom_name'])


if __name__ == "__main__":
    unittest.main()
